# Copyright (c) 2023 promptulate
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# Copyright Owner: Zeeland
# GitHub Link: https://github.com/Undertone0809/
# Project Link: https://github.com/Undertone0809/promptulate
# Contact Email: zeeland4work@gmail.com


from abc import ABC, abstractmethod
from typing import Optional, TypeVar, Union

from promptulate.hook import Hook, HookTable
from promptulate.pydantic_v1 import BaseModel
from promptulate.schema import (
    AssistantMessage,
    BaseMessage,
    LLMType,
    MessageSet,
)

T = TypeVar("T", bound=BaseModel)


class BaseLLM(BaseModel, ABC):
    llm_type: Union[str, LLMType] = "custom"

    class Config:
        """Configuration for this pydantic object."""

        extra = "allow"
        arbitrary_types_allowed = True

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        if "hooks" in kwargs and kwargs["hooks"]:
            for hook in kwargs["hooks"]:
                Hook.mount_instance_hook(hook, self)
        Hook.call_hook(HookTable.ON_LLM_CREATE, self, **kwargs)

    def predict(self, messages: MessageSet, *args, **kwargs) -> AssistantMessage:
        """llm generate prompt"""
        Hook.call_hook(HookTable.ON_LLM_START, self, messages, *args, **kwargs)
        result = self._predict(messages, *args, **kwargs)
        if isinstance(result, AssistantMessage):
            Hook.call_hook(HookTable.ON_LLM_RESULT, self, result=result.content)

        return result

    @abstractmethod
    def _predict(
        self, messages: MessageSet, *args, **kwargs
    ) -> Optional[type(BaseMessage)]:
        """Run the llm, implemented through subclass."""
        raise NotImplementedError()

    def __call__(self, instruction: str, *args, **kwargs) -> str:
        """Invoke the llm to generate a string response, which is a simplified version
        of predict.

        Args:
            instruction(str): The instruction to be processed.

        Returns:
            str: The response generated by the llm.
        """
        messages = MessageSet.from_listdict_data(
            [
                {"role": "system", "content": "You are a helpful assistant."},
                {"role": "user", "content": instruction},
            ]
        )
        return self.predict(messages, **kwargs).content



from promptulate.agents.tool_agent.agent import ToolAgent
from promptulate.llms.base import BaseLLM
from promptulate.tools.duckduckgo.tools import ddg_websearch


class WebAgent(ToolAgent):
    def __init__(self, llm: BaseLLM = None, *args, **kwargs):
        super().__init__(tools=[ddg_websearch], llm=llm, *args, **kwargs)



from promptulate.tools.semantic_scholar.tools import (
    SemanticScholarCitationTool,
    SemanticScholarQueryTool,
    SemanticScholarReferenceTool,
)

__all__ = [
    "SemanticScholarQueryTool",
    "SemanticScholarReferenceTool",
    "SemanticScholarCitationTool",
]



from promptulate.agents.assistant_agent.schema import (
    AgentPlanResponse,
    Plan,
    plan_examples,
)
from promptulate.agents.planner.prompt import PLAN_SYS_PROMPT
from promptulate.chat import AIChat
from promptulate.llms.base import BaseLLM
from promptulate.output_formatter import OutputFormatter


class Planner:
    def __init__(self, llm: BaseLLM, system_prompt: str):
        self.ai = AIChat(custom_llm=llm)
        self.system_prompt: str = system_prompt

    def run(self, instruction: str) -> Plan:
        """Generate a plan for the user target."""
        system_prompt = PLAN_SYS_PROMPT + instruction
        formatter = OutputFormatter(AgentPlanResponse, plan_examples)
        system_prompt += formatter.get_formatted_instructions()

        resp: str = self.ai.run(system_prompt)
        plan_resp: AgentPlanResponse = formatter.formatting_result(resp)

        return Plan.parse_obj({**plan_resp.dict(), "next_task_id": 1})



from typing import Callable, List, Optional, Tuple, Union

from typing_extensions import Literal

from promptulate.pydantic_v1 import BaseModel
from promptulate.utils.core_utils import generate_unique_id
from promptulate.utils.logger import logger

HOOK_TYPE = Literal["component", "instance"]
# Hook component type
COMPONENT_TYPE = Literal["Tool", "llm", "Agent"]


class HookTable:
    ON_LLM_CREATE = ("llm", "on_llm_create")
    ON_LLM_START = ("llm", "on_llm_start")
    ON_LLM_RESULT = ("llm", "on_llm_result")

    ON_TOOL_CREATE = ("Tool", "on_tool_create")
    ON_TOOL_START = ("Tool", "on_tool_start")
    ON_TOOL_RESULT = ("Tool", "on_tool_result")

    ON_AGENT_CREATE = ("Agent", "on_agent_create")
    ON_AGENT_PLAN = ("Agent", "on_agent_plan")
    ON_AGENT_REVISE_PLAN = ("Agent", "on_agent_revise_plan")
    ON_AGENT_START = ("Agent", "on_agent_start")
    ON_AGENT_ACTION = ("Agent", "on_agent_action")
    ON_AGENT_OBSERVATION = ("Agent", "on_agent_observation")
    ON_AGENT_RESULT = ("Agent", "on_agent_result")


class BaseHookSchema(BaseModel):
    hook_name: str
    """Hook lifecycle."""
    callback: Callable
    """Callback function of hook."""
    component_type: COMPONENT_TYPE
    """Component type for hook."""

    def __str__(self):
        return f"<HookSchema> hook_name: {self.hook_name} callback: {self.callback} component_type: {self.component_type}"  # noqa


ComponentHookSchema = BaseHookSchema


class InstanceHookSchema(BaseHookSchema):
    mounted_obj: Optional[object]
    """Mounted obj call hooks. When hook is registered as None, it will only be assigned
    a value when the component is instantiated."""
    unique_id: str

    def __str__(self):
        return (
            f"<HookSchema> hook_name: {self.hook_name} callback: {self.callback} component_type: "  # noqa
            f"{self.component_type} mounted_obj: {self.mounted_obj.__class__.__name__}"
        )


def _hook_decorator(
    hook_table: Tuple,
    hook_type: HOOK_TYPE,
    fn: Callable,
):
    """Hook decorator wrapper for HookMixin"""
    unique_id = generate_unique_id()

    if hook_type == "instance":
        Hook.registry_instance_hook(hook_table, fn, unique_id)
    elif hook_type == "component":
        Hook.registry_component_hook(hook_table, fn)
    else:
        raise ValueError("hook type error, `instance` or `component` are specified.")

    def wrapper(*args, **kwargs):
        return fn(*args, **kwargs)

    wrapper.__unique_id__ = unique_id

    return wrapper


class ToolHookMixin:
    @staticmethod
    def on_tool_create(hook_type: HOOK_TYPE):
        def decorator(fn):
            return _hook_decorator(HookTable.ON_TOOL_CREATE, hook_type, fn)

        return decorator

    @staticmethod
    def on_tool_start(hook_type: HOOK_TYPE):
        def decorator(fn):
            return _hook_decorator(HookTable.ON_TOOL_START, hook_type, fn)

        return decorator

    @staticmethod
    def on_tool_result(hook_type: HOOK_TYPE):
        def decorator(fn):
            return _hook_decorator(HookTable.ON_TOOL_RESULT, hook_type, fn)

        return decorator


class AgentHookMixin:
    @staticmethod
    def on_agent_create(hook_type: HOOK_TYPE):
        def decorator(fn):
            return _hook_decorator(HookTable.ON_AGENT_CREATE, hook_type, fn)

        return decorator

    @staticmethod
    def on_agent_start(hook_type: HOOK_TYPE):
        def decorator(fn):
            return _hook_decorator(HookTable.ON_AGENT_START, hook_type, fn)

        return decorator

    @staticmethod
    def on_agent_plan(hook_type: HOOK_TYPE):
        def decorator(fn):
            return _hook_decorator(HookTable.ON_AGENT_PLAN, hook_type, fn)

        return decorator

    @staticmethod
    def on_agent_revise_plan(hook_type: HOOK_TYPE):
        def decorator(fn):
            return _hook_decorator(HookTable.ON_AGENT_REVISE_PLAN, hook_type, fn)

        return decorator

    @staticmethod
    def on_agent_action(hook_type: HOOK_TYPE):
        def decorator(fn):
            return _hook_decorator(HookTable.ON_AGENT_ACTION, hook_type, fn)

        return decorator

    @staticmethod
    def on_agent_observation(hook_type: HOOK_TYPE):
        def decorator(fn):
            return _hook_decorator(HookTable.ON_AGENT_OBSERVATION, hook_type, fn)

        return decorator

    @staticmethod
    def on_agent_result(hook_type: HOOK_TYPE):
        def decorator(fn):
            return _hook_decorator(HookTable.ON_AGENT_RESULT, hook_type, fn)

        return decorator


class LLMHookMixin:
    @staticmethod
    def on_llm_create(hook_type: HOOK_TYPE):
        def decorator(fn):
            return _hook_decorator(HookTable.ON_LLM_CREATE, hook_type, fn)

        return decorator

    @staticmethod
    def on_llm_start(hook_type: HOOK_TYPE):
        def decorator(fn):
            return _hook_decorator(HookTable.ON_LLM_START, hook_type, fn)

        return decorator

    @staticmethod
    def on_llm_result(hook_type: HOOK_TYPE):
        def decorator(fn):
            return _hook_decorator(HookTable.ON_LLM_RESULT, hook_type, fn)

        return decorator


class Hook(ToolHookMixin, AgentHookMixin, LLMHookMixin):
    component_hook_store: List[ComponentHookSchema] = []
    instance_hook_store: List[InstanceHookSchema] = []
    unmounted_hook_store: List[InstanceHookSchema] = []

    @classmethod
    def registry_hook(
        cls,
        hook_table: Tuple,
        callbacks: Union[Callable, List[Callable]],
        hook_type: HOOK_TYPE,
    ):
        """Registry component or instance hook to the hook center for unified
        management."""
        if isinstance(callbacks, Callable):
            callbacks = [callbacks]

        for callback in callbacks:
            if hook_type == "component":
                cls.registry_component_hook(hook_table, callback)
            elif hook_type == "instance":
                cls.registry_instance_hook(hook_table, callback)

    @classmethod
    def registry_component_hook(cls, hook_table: Tuple, callback: Callable):
        hook = ComponentHookSchema(
            hook_name=hook_table[1],
            callback=callback,
            component_type=hook_table[0],
        )
        logger.debug(f"[pne hook registry] instance hook: {hook}")
        cls.component_hook_store.append(hook)

    @classmethod
    def registry_instance_hook(
        cls,
        hook_table: Tuple,
        callback: Callable,
        unique_id: Optional[str] = None,
    ):
        """Registry instance hook firstly, it still needs to be mounted by
        `mount_instance_hook`

        Args:
            hook_table: hook table
            callback: hook callback
            unique_id: a temporary variable
        """
        if not unique_id and not getattr(callback, "__unique_id__", None):
            unique_id = generate_unique_id()
            callback.__unique_id__ = unique_id

        hook = InstanceHookSchema(
            hook_name=hook_table[1],
            callback=callback,
            component_type=hook_table[0],
            unique_id=unique_id,
        )
        logger.debug(f"[pne hook registry] instance hook: {hook}")
        cls.unmounted_hook_store.append(hook)

    @classmethod
    def unregister_hook(cls, callback: Callable):
        for hook in cls.component_hook_store:
            if id(hook.callback) == id(callback):
                cls.component_hook_store.remove(hook)
        for hook in cls.instance_hook_store:
            if id(hook.callback) == id(callback):
                cls.instance_hook_store.remove(hook)

    @classmethod
    def mount_instance_hook(cls, hook_callback: Callable, mounted_obj: object):
        for unmounted_hook in Hook.unmounted_hook_store:
            if (
                getattr(hook_callback, "__unique_id__", None)
                and unmounted_hook.unique_id == hook_callback.__unique_id__
            ):
                hook = unmounted_hook.copy()
                hook.mounted_obj = mounted_obj
                Hook.instance_hook_store.append(hook)
                Hook.unmounted_hook_store.remove(unmounted_hook)
                logger.debug(f"[pne hook mount] {hook}")
                return
        raise ValueError(
            f"Could not mount function <{hook_callback}> to {mounted_obj.__class__.__name__}. "  # noqa
            "You may have used component Hook, but component Hook cannot be mounted to an instance."  # noqa
        )

    @classmethod
    def get_hooks(cls, hook_name: str, mounted_obj: object) -> List[BaseHookSchema]:
        """Get relevant component hooks and instance hooks for mounted_obj from hook
        store."""
        return [
            *Hook._get_component_hooks(hook_name),
            *Hook._get_instance_hooks(hook_name, mounted_obj),
        ]

    @classmethod
    def _get_component_hooks(cls, hook_name: str) -> List[BaseHookSchema]:
        hooks: List[BaseHookSchema] = []
        for hook in cls.component_hook_store:
            if hook.hook_name == hook_name:
                hooks.append(hook)
        return hooks

    @classmethod
    def _get_instance_hooks(
        cls, hook_name: str, mounted_obj: object
    ) -> List[BaseHookSchema]:
        hooks: List[BaseHookSchema] = []
        for hook in cls.instance_hook_store:
            if hook.hook_name == hook_name and hook.mounted_obj is mounted_obj:
                hooks.append(hook)
        return hooks

    @classmethod
    def call_hook(cls, hook_table: Tuple, mounted_obj: object, *args, **kwargs):
        """Invoke the specified hook when specifying the component lifecycle."""
        hooks = cls.get_hooks(hook_table[1], mounted_obj)

        if hooks:
            logger.debug(
                f"[pne hook] hooks <{hook_table}> mounted_obj <{mounted_obj.__class__.__name__}> call hook: <{hooks}>"  # noqa
            )
            for hook in hooks:
                hook.callback(*args, **kwargs)



# Copyright (c) 2023 promptulate
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# Copyright Owner: Zeeland
# GitHub Link: https://github.com/Undertone0809/
# Project Link: https://github.com/Undertone0809/promptulate
# Contact Email: zeeland@foxmail.com

import os
from typing import Optional

import requests
from dotenv import load_dotenv

from promptulate.hook.stdout_hook import StdOutHook
from promptulate.utils.core_utils import get_cache
from promptulate.utils.logger import logger
from promptulate.utils.openai_key_pool import OpenAIKeyPool
from promptulate.utils.singleton import Singleton

PROXY_MODE = ["off", "custom", "promptulate"]
load_dotenv()


def set_enable_cache(value: bool):
    """Caching is enabled by default. Disabling caching is not recommended."""
    Config().enable_cache = value


def turn_off_stdout_hook():
    Config().turn_off_stdout_hook()


class Config(metaclass=Singleton):
    """
    Configuration class for managing application settings.

    Attributes:
        enable_cache (bool): Flag indicating whether caching is enabled.
        _proxy_mode (str): The current proxy mode.
        _proxies (Optional[dict]): Dictionary of proxy settings.
        openai_chat_api_url (str): URL for OpenAI chat API.
        openai_completion_api_url (str): URL for OpenAI completion API.
        openai_proxy_url (str): URL for OpenAI proxy.
        ernie_bot_url (str): URL for Ernie bot API.
        ernie_bot_turbo_url (str): URL for Ernie bot turbo API.
        ernie_bot_token_url (str): URL for Ernie bot token API.
        ernie_bot_4_url (str): URL for Ernie bot 4 API.
        ernie_embedding_v1_url (str): URL for Ernie embedding v1 API.
        key_default_retry_times (int): Default number of key retry times.
        enable_stdout_hook (bool): Flag indicating whether stdout hook is enabled.
    """

    def __init__(self):
        logger.info("[pne config] Config initialization")
        self.enable_cache: bool = True
        self._proxy_mode: str = PROXY_MODE[0]
        self._proxies: Optional[dict] = None

        self.openai_chat_api_url = "https://api.openai.com/v1/chat/completions"
        self.openai_completion_api_url = "https://api.openai.com/v1/completions"
        self.openai_proxy_url = "https://chatgpt-api.shn.hk/v1/"  # FREE API

        self.ernie_bot_url = "https://aip.baidubce.com/rpc/2.0/ai_custom/v1/wenxinworkshop/chat/completions"
        self.ernie_bot_turbo_url = "https://aip.baidubce.com/rpc/2.0/ai_custom/v1/wenxinworkshop/chat/eb-instant"
        self.ernie_bot_token_url = "https://aip.baidubce.com/oauth/2.0/token"
        self.ernie_bot_4_url = "https://aip.baidubce.com/rpc/2.0/ai_custom/v1/wenxinworkshop/chat/completions_pro"
        self.ernie_embedding_v1_url = "https://aip.baidubce.com/rpc/2.0/ai_custom/v1/wenxinworkshop/embeddings/embedding-v1"

        self.zhipu_model_url = "https://open.bigmodel.cn/api/paas/v4/chat/completions"

        self.key_default_retry_times = 5
        """If llm(like OpenAI) unable to obtain data, retry request until the data is obtained."""  # noqa: E501
        self.enable_stdout_hook = True

        if self.enable_stdout_hook:
            StdOutHook.registry_stdout_hooks()

    def turn_off_stdout_hook(self):
        if self.enable_stdout_hook:
            self.enable_stdout_hook = False
            StdOutHook.unregister_stdout_hooks()

    def get_openai_api_key(self, model: str = "gpt-3.5-turbo") -> str:
        """Get openai key from KeyPool and environment variable. The priority of
        obtaining the key: environment variable, cache of OpenAIKeyPool, cache of
        OEPNAI_API_KEY."""
        if "OPENAI_API_KEY" in os.environ.keys():
            if self.enable_cache:
                get_cache()["OPENAI_API_KEY"] = os.getenv("OPENAI_API_KEY")
            return os.getenv("OPENAI_API_KEY")

        if self.enable_cache:
            openai_key_pool: OpenAIKeyPool = OpenAIKeyPool()
            key = openai_key_pool.get(model)
            if key:
                return key

        if self.enable_cache and "OPENAI_API_KEY" in get_cache():
            return get_cache()["OPENAI_API_KEY"]

        raise ValueError("OPENAI API key is not provided. Please set your key.")

    def get_ernie_api_key(self) -> str:
        if "ERNIE_API_KEY" in os.environ.keys():
            if self.enable_cache:
                get_cache()["ERNIE_API_KEY"] = os.getenv("ERNIE_API_KEY")
            return os.getenv("ERNIE_API_KEY")
        if self.enable_cache and "ERNIE_API_KEY" in get_cache():
            return get_cache()["ERNIE_API_KEY"]
        raise ValueError("ERNIE_API_KEY is not provided. Please set your key.")

    @property
    def get_ernie_api_secret(self) -> str:
        if "ERNIE_API_SECRET" in os.environ.keys():
            if self.enable_cache:
                get_cache()["ERNIE_API_SECRET"] = os.getenv("ERNIE_API_SECRET")
            return os.getenv("ERNIE_API_SECRET")
        if self.enable_cache and "ERNIE_API_SECRET" in get_cache():
            return get_cache()["ERNIE_API_SECRET"]
        raise ValueError("ERNIE_API_SECRET is not provided. Please set your secret.")

    def get_zhipuai_api_key(self) -> str:
        if "ZHIPUAI_API_KEY" in os.environ.keys():
            if self.enable_cache:
                get_cache()["ZHIPUAI_API_KEY"] = os.getenv("ZHIPUAI_API_KEY")
            return os.getenv("ZHIPUAI_API_KEY")
        if self.enable_cache and "ZHIPUAI_API_KEY" in get_cache():
            return get_cache()["ZHIPUAI_API_KEY"]
        raise ValueError("ZHIPUAI_API_KEY is not provided. Please set your key.")

    def get_qianfan_ak(self):
        if "QIANFAN_ACCESS_KEY" in os.environ.keys():
            if self.enable_cache:
                get_cache()["QIANFAN_ACCESS_KEY"] = os.getenv("QIANFAN_ACCESS_KEY")
            return os.getenv("QIANFAN_ACCESS_KEY")
        if self.enable_cache and "QIANFAN_ACCESS_KEY" in get_cache():
            return get_cache()["QIANFAN_ACCESS_KEY"]
        raise ValueError("QIANFAN_ACCESS_KEY is not provided. Please set your key.")

    def get_qianfan_sk(self):
        if "QIANFAN_SECRET_KEY" in os.environ.keys():
            if self.enable_cache:
                get_cache()["QIANFAN_SECRET_KEY"] = os.getenv("QIANFAN_SECRET_KEY")
            return os.getenv("QIANFAN_SECRET_KEY")
        if self.enable_cache and "QIANFAN_SECRET_KEY" in get_cache():
            return get_cache()["QIANFAN_SECRET_KEY"]
        raise ValueError("QIANFAN_SECRET_KEY is not provided. Please set your secret.")

    def get_ernie_token(self) -> str:
        url = self.ernie_bot_token_url
        params = {
            "grant_type": "client_credentials",
            "client_id": self.get_ernie_api_key(),
            "client_secret": self.get_ernie_api_secret,
        }
        return str(requests.post(url, params=params).json().get("access_token"))

    def get_key_retry_times(self, model: str) -> int:
        if self.enable_cache:
            openai_key_pool: OpenAIKeyPool = OpenAIKeyPool()
            return openai_key_pool.get_num(model)
        return self.key_default_retry_times

    @property
    def proxy_mode(self) -> str:
        if self.enable_cache and "PROXY_MODE" in get_cache():
            return get_cache()["PROXY_MODE"]
        return self._proxy_mode

    @proxy_mode.setter
    def proxy_mode(self, value):
        self._proxy_mode = value
        if self.enable_cache:
            get_cache()["PROXY_MODE"] = value

    @property
    def proxies(self) -> Optional[dict]:
        if self.enable_cache and "PROXIES" in get_cache():
            return get_cache()["PROXIES"] if self.proxy_mode == "custom" else None
        return self._proxies

    @proxies.setter
    def proxies(self, value):
        self._proxies = value
        if self.enable_cache:
            get_cache()["PROXIES"] = value

    @property
    def openai_chat_request_url(self) -> str:
        if self.proxy_mode == PROXY_MODE[2]:
            self.proxies = None
            return self.openai_proxy_url
        return self.openai_chat_api_url

    @property
    def openai_completion_request_url(self) -> str:
        if self.proxy_mode == PROXY_MODE[2]:
            self.proxies = None
            return f"{self.openai_proxy_url}completions"
        return self.openai_completion_api_url

    def set_proxy_mode(self, mode: str, proxies: Optional[dict] = None):
        self.proxy_mode = mode
        self.proxies = proxies
        logger.info(f"[pne] proxy mode: {mode}, proxies: {proxies}")


pne_config = Config()



from promptulate.utils import StringTemplate

PLAN_SYSTEM_PROMPT = """
For the given objective, come up with a simple step by step plan.
This plan should involve individual tasks, that if executed correctly will yield the correct answer. Do not add any superfluous steps.
The result of the final step should be the final answer. Make sure that each step has all the information needed - do not skip steps.

{{user_target}}
"""  # noqa

REVISE_SYSTEM_PROMPT = """
As a powerfule agent, your job is to help the user achieve their goal. This is user instruction: {{user_target}}

## Workflows
Currently, you are working on a job that has been planned. The plan execution details / job description is provided. Please review the plan and revise the plan according to the execution details to meet the goal. You should follow the following rules:
1. If you think any task is DONE already, mark the task status as DONE.
2. If the task execution result does not meet the target, mark the task status as ERROR.
3. If you think current plan is missing tasks to meet the goal, you can add new tasks.
4. Some of the task may be invalid, mark the task status as DISCARDED.
5. Pay attention to the status of each of the tasks.
6. If you think the plan is complete, next task id should be None.

Your original plan was this:
{{original_plan}}

You have currently done the follow steps:
{{past_steps}}

## Task
Update your plan accordingly. If no more steps are needed and you can return to the user, then respond with that. Otherwise, fill out the plan. Only add steps to the plan that still NEED to be done. Do not return previously done steps as part of the plan.
You need to output the updated plan and next task id.

## Constraints 
- Next task id must be searchable in the current task list.
"""  # noqa

PLAN_SYSTEM_PROMPT_TMP = StringTemplate(PLAN_SYSTEM_PROMPT, "jinja2")
REVIEW_SYSTEM_PROMPT_TMP = StringTemplate(REVISE_SYSTEM_PROMPT, "jinja2")



import os
from typing import List, Literal, Optional

from promptulate.tools.base import BaseToolKit, ToolTypes
from promptulate.tools.file.tools import (
    AppendFileTool,
    CopyFileTool,
    DeleteFileTool,
    ListDirectoryTool,
    MoveFileTool,
    ReadFileTool,
    WriteFileTool,
)

FileToolType = Literal["write", "append", "read", "delete", "list", "copy", "move"]
TOOL_MAPPER = {
    "write": WriteFileTool,
    "append": AppendFileTool,
    "read": ReadFileTool,
    "delete": DeleteFileTool,
    "list": ListDirectoryTool,
    "copy": CopyFileTool,
    "move": MoveFileTool,
}


class FileToolKit(BaseToolKit):
    """File ToolKit

    Args:
        root_dir: The root directory of the file tool.
        modes(Option): The modes of the file tool. Default is None.

    Returns:
        The instance object of the corresponding tool
    """

    def __init__(
        self,
        *,
        root_dir: Optional[str] = None,
        modes: Optional[List[FileToolType]] = None,
    ) -> None:
        self.root_dir = root_dir or os.getcwd()
        self.modes = modes or []

        for mode in self.modes:
            if mode not in TOOL_MAPPER.keys():
                raise ValueError(
                    f"{mode} does not exist.\n"
                    f"Please select from {list(TOOL_MAPPER.keys())}"
                )

    def get_tools(self) -> List[ToolTypes]:
        if self.modes:
            return [TOOL_MAPPER[mode](self.root_dir) for mode in self.modes]
        return [tool(self.root_dir) for tool in TOOL_MAPPER.values()]



import json
import re
from typing import Any, Dict, List, Optional, Type, TypeVar, Union

from promptulate.error import OutputParserError
from promptulate.output_formatter.prompt import OUTPUT_FORMAT
from promptulate.pydantic_v1 import BaseModel

T = TypeVar("T", bound=BaseModel)


def _get_schema(pydantic_obj: Type[BaseModel]) -> dict:
    """Get reduced schema from pydantic object.

    Args:
        pydantic_obj: The pydantic object to get schema from.

    Returns:
        Dict: The reduced schema, which is without title and type fields.
    """

    # Remove useless fields.
    # Compatibility with Pydantic v2 model_json_schema method.
    if hasattr(pydantic_obj, "model_json_schema"):
        return pydantic_obj.model_json_schema()
    else:
        reduced_schema = pydantic_obj.schema()

    if "title" in reduced_schema:
        del reduced_schema["title"]
    if "type" in reduced_schema:
        del reduced_schema["type"]
    for k, v in reduced_schema["properties"].items():
        if "title" in v:
            del v["title"]

    return reduced_schema


class OutputFormatter:
    """
    Class for formatting the output of a Pydantic object.

    This class provides methods for getting formatted instructions and formatting the
    result of a Pydantic object.
    """

    def __init__(
        self,
        pydantic_obj: Type[BaseModel],
        examples: Optional[List[BaseModel]] = None,
        **kwargs,
    ):
        """
        Initialize the OutputFormatter class.

        Args:
            pydantic_obj (type(BaseModel)): The Pydantic object to format.
            examples (List[BaseModel], optional): Examples of the Pydantic object.
        """
        super().__init__(**kwargs)

        self.pydantic_obj: Type[BaseModel] = pydantic_obj
        self.examples: Optional[List[BaseModel]] = examples

    def get_formatted_instructions(self) -> str:
        return get_formatted_instructions(self.pydantic_obj, self.examples)

    def formatting_result(self, llm_output: str) -> T:
        return formatting_result(self.pydantic_obj, llm_output)


def get_formatted_instructions(
    json_schema: Union[Type[BaseModel], Dict[str, Any]],
    examples: List[BaseModel] = None,
) -> str:
    """
    Get formatted instructions for a JSON schema or Pydantic object.

    Args:
        json_schema (Union[Dict, type(BaseModel)]): The JSON schema or Pydantic object
            to get instructions for.
        examples (List[BaseModel], optional): Examples of the JSON schema or Pydantic
            object. Defaults to None.

    Returns:
        str: The formatted instructions.
    """
    # If a Pydantic model is passed, extract the schema from it.
    if not isinstance(json_schema, dict):
        json_schema: dict = _get_schema(json_schema)

    # Ensure json with double quotes.
    schema_str = json.dumps(json_schema)

    instructions: str = OUTPUT_FORMAT.format(schema=schema_str)
    if examples:
        instructions += "\nExamples:\n"
        for example in examples:
            example_str = json.dumps(example.dict())
            instructions += f"{example_str}\n"
    return instructions


def formatting_result(pydantic_obj: type(BaseModel), llm_output: str) -> T:
    """
    Parse llm_output and instantiate the result as provided pydantic_obj.

    Args:
        pydantic_obj (type(BaseModel)): The Pydantic object to instantiate the result
            as.
        llm_output (str): The output to parse.

    Returns:
        T: The instantiated result.

    Raises:
        OutputParserError: If the output cannot be parsed.
    """
    try:
        match = re.search(
            r"\{.*\}", llm_output.strip(), re.MULTILINE | re.IGNORECASE | re.DOTALL
        )
        json_str = match.group() if match else ""
        json_object = json.loads(json_str, strict=False)

        # Compatibility with Pydantic v2 model_validate method.
        if hasattr(pydantic_obj, "model_validate"):
            return pydantic_obj.model_validate(json_object)
        else:
            return pydantic_obj.parse_obj(json_object)

    except Exception as e:
        name = pydantic_obj.__name__
        msg = f"Failed to parse {name} from completion {llm_output}. Got: {e}"
        raise OutputParserError(msg, llm_output)



from typing import Callable

from promptulate.tools import Tool
from promptulate.utils.color_print import print_text


def _print_func(llm_question: str) -> None:
    """Default way to show llm question when llm using HumanFeedBackTool."""
    print_text(f"[Agent ask] {llm_question}", "blue")


class HumanFeedBackTool(Tool):
    """A tool for human feedback"""

    name: str = "human_feedback"
    description: str = (
        "Human feedback tools are used to collect human feedback information."
        "Please only use this tool in situations where relevant contextual information"
        "is lacking or reasoning cannot continue. Please enter the content you wish for"
        "human feedback and interaction, but do not ask for knowledge or let humans reason."  # noqa
    )

    def __init__(
        self,
        output_func: Callable[[str], None] = _print_func,
        input_func: Callable = input,
        *args,
        **kwargs,
    ):
        super().__init__(*args, **kwargs)
        self.output_func = output_func
        self.input_func = input_func

    def _run(self, content: str, *args, **kwargs) -> str:
        self.output_func(content)
        return self.input_func()



from typing import Any, Dict, Optional

from promptulate.llms._litellm import LiteLLM
from promptulate.llms.base import BaseLLM


class LLMFactory:
    @classmethod
    def build(
        cls, model_name: str, *, model_config: Optional[Dict[str, Any]] = None, **kwargs
    ) -> BaseLLM:
        model_config = model_config or {}

        try:
            provider, model_id = model_name.split("/")

            if provider == "zhipu":
                from promptulate.llms import ZhiPu

                return ZhiPu(model=model_id, model_config=model_config, **kwargs)
            elif provider == "qianfan":
                from promptulate.llms import QianFan

                return QianFan(model=model_id, model_config=model_config, **kwargs)
        except ValueError:
            pass

        return LiteLLM(model=model_name, model_config=model_config, **kwargs)



import time
from typing import Dict, List, Optional, Union

from promptulate.tools.base import Tool
from promptulate.tools.duckduckgo.api_wrapper import DuckDuckGoSearchAPIWrapper
from promptulate.utils.core_utils import listdict_to_string


class DuckDuckGoTool(Tool):
    """Tool that adds the capability to query the DuckDuckGo search API."""

    name: str = "ddg-search"
    description: str = (
        "A wrapper around DuckDuckgo Web Search. "
        "Useful for when you need to answer questions about current events. "
        "Args : keyword(str)"
        "Input should be a search query."
    )

    def __init__(self, max_retry: Optional[int] = None, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.max_retry: int = max_retry or 5
        self.api_wrapper = DuckDuckGoSearchAPIWrapper()

    def _run(self, keyword: str, **kwargs) -> Union[str, List[str]]:
        """Run duckduckgo search and get search result.

        Args:
            keyword: query keyword
            **kwargs:
                result_type(Optional[str]) - default return string type data. Return
                List[str] type data.
                if you pass result_type="original"

        Returns:
            default is string. Return List[str] type data if you pass
            result_type="original"
        """
        from duckduckgo_search.exceptions import RateLimitException

        attempt = 0
        while attempt < self.max_retry:
            try:
                result = self.api_wrapper.query(keyword, **kwargs)
                if "result_type" in kwargs and kwargs["result_type"] == "original":
                    return result
                return " ".join(result)
            except RateLimitException as e:
                attempt += 1

                if attempt >= self.max_retry:
                    raise e

                time.sleep(1.3**attempt)  # Exponential backoff
            except Exception as e:
                raise e


class DuckDuckGoReferenceTool(Tool):
    """Tool that adds the capability to query the DuckDuckGo search API. Compared to
    DuckDuckGoTool, this tool can return references information like href, title.
    """

    name: str = "ddg-search-with-references"
    description: str = (
        "A wrapper around DuckDuckGo Search, it returns query result and references url"
        "Useful for when you need to answer questions about current events. "
        "Input should be a search query."
    )

    def __init__(self, max_retry: Optional[int] = None, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.max_retry: int = max_retry or 5
        self.api_wrapper = DuckDuckGoSearchAPIWrapper()

    def _run(self, keyword: str, **kwargs) -> Union[str, List[Dict[str, str]]]:
        """Run duckduckgo search and get search result with href, snippet, title.

        Args:
            keyword: query keyword
            **kwargs:
                result_type(Optional[str]) - default return string type data. Return
                List[str] type data.
                if you pass result_type="original"

        Returns:
            default is string. Return List[Dict[str, str]] type data if you pass
            result_type="original"
        """
        result = self.api_wrapper.query_by_formatted_results(keyword, **kwargs)
        if "return_type" in kwargs and kwargs["return_type"] == "original":
            return result
        return listdict_to_string(result, item_suffix="\n")


def ddg_websearch(query: str) -> str:
    """Run duckduckgo search and get search result.

    Args:
        query: query keyword
    """
    tool = DuckDuckGoTool()
    return tool.run(query)



import json
import time
from typing import Dict, List, Optional

import requests

from promptulate.error import NetWorkError
from promptulate.pydantic_v1 import BaseModel
from promptulate.utils.logger import logger


class SemanticScholarAPIWrapper(BaseModel):
    """https://api.semanticscholar.org/api-docs/graph#tag/Paper-Data/"""

    BASE_API_URL: str = "https://api.semanticscholar.org/graph/v1"
    BASE_OFFICIAL_URL: str = "https://www.semanticscholar.org"
    max_result: int = 10
    current_result: Optional[List[Dict]] = None
    paper_query_params: Dict[str, str] = {"fields": "title,url,abstract"}

    def _get_string_fields(self) -> str:
        return f"&fields={self.paper_query_params['fields']}&"

    def get_paper(
        self, query: str, max_result: Optional[int] = None, **kwargs
    ) -> List[Dict]:
        """This method can obtain a list of relevant papers based on your query.

        Args:
            max_result: num of max result
            query: the papers you want to query
            **kwargs:
                specified_fields(Optional[List[str]]): filter specified field to return.
                For example, you can return the ["title", "url", "abstract"] fields from
                each arxiv query result.
                You can see the detail fields as follows:
                https://api.semanticscholar.org/api-docs/graph#tag/Paper-Data/operation/post_graph_get_papers

        Returns:
            Return List[Dict] data, the default detail fields is as follows.
            - id (str): semantic id
            - authorsYear (str): author and publication year
            - title (str): paper title
            - url (str): paper semantic scholar url
            If you want to get more detail fields, please use `specified_fields`
        """

        def get_detail():
            """
            Get more paper attributions for specified_fields.
            ref: https://api.semanticscholar.org/api-docs/graph#tag/Paper-Data/operation/post_graph_get_papers
            """
            paper_ids = list(map(lambda p: p["id"], self.current_result))
            fields: List[str] = kwargs["specified_fields"]
            params: Dict[str, str] = {"fields": ",".join(fields)}

            resp = requests.post(
                f"{self.BASE_API_URL}/paper/batch",
                params=params,
                json={"ids": paper_ids},
            )
            if response.status_code != 200:
                raise NetWorkError("SemanticScholarAPIWrapper.get_paper.get_detail")

            detail_result = resp.json()
            if isinstance(detail_result, dict):
                raise NetWorkError(
                    "SemanticScholarAPIWrapper.get_paper.get_detail",
                    reason=detail_result["message"],
                )

            detail_result = resp.json()

            time.sleep(0.1)

            for original in self.current_result:
                for detail_item in detail_result:
                    if detail_item["paperId"] == original["id"]:
                        original.update(detail_item)
            return

        if not max_result:
            max_result = self.max_result

        query = "%20".join(query.split(" "))
        url = f"{self.BASE_API_URL}/paper/autocomplete?query={query}&limit={max_result}"
        response = requests.get(url)
        if response.status_code == 200:
            self.current_result = response.json()["matches"]

            if len(self.current_result) == 0:
                logger.debug("[pne] semantic scholar return none")
                return []

            for item in self.current_result:
                item["url"] = self._get_url(item["id"])

            if "specified_fields" in kwargs:
                get_detail()
            logger.debug(
                f"[pne] semantic scholar result {json.dumps(self.current_result)}"
            )
            return self.current_result
        raise NetWorkError("semantic scholar query")

    def _get_url(self, id: str) -> str:
        """Get paper url from paper id."""
        return f"{self.BASE_OFFICIAL_URL}/paper/{id}"

    def get_references(self, query: str, max_result: int = 500, **kwargs) -> List[Dict]:
        """Used to get references of specified paper.

        Args:
            max_result: num of max result
            query: the paper you want to query

        Returns:
            Return List[Dict] data, the default detail fields is as follows.
            - id (str): semantic id
            - authorsYear (str): author and publication year
            - title (str): paper title
            - url (str): paper semantic scholar url
        """
        papers = self.get_paper(query)
        if len(papers) == 0:
            return []

        paper_id: str = papers[0]["id"]
        url = f"{self.BASE_API_URL}/paper/{paper_id}/references?offset=1&limit={max_result}"  # noqa
        response = requests.get(url)
        if response.status_code == 200:
            res_data = response.json()["data"]
            final_result: List[Dict] = []
            for i, item in enumerate(res_data):
                if not item["citedPaper"]["paperId"]:
                    continue

                item["citedPaper"]["url"] = self._get_url(item["citedPaper"]["paperId"])
                item["citedPaper"]["id"] = item["citedPaper"]["paperId"]
                del item["citedPaper"]["paperId"]
                final_result.append(item["citedPaper"])

            logger.debug(f"[pne semantic scholar result] {json.dumps(final_result)}")
            return final_result
        raise NetWorkError("semantic scholar")

    def get_citations(self, query: str, max_result: Optional[int] = None):
        """Used to get citation of specified paper.

        Args:
            max_result: num of max result
            query: the paper you want to query

        Returns:
            Return List[Dict] data, the default detail fields is as follows.
            - id (str): semantic id
            - abstract (str): paper abstract
            - title (str): paper title
            - url (str): paper semantic scholar url
        """
        papers = self.get_paper(query)
        if len(papers) == 0:
            return []
        if not max_result:
            max_result = self.max_result

        paper_id: str = papers[0]["id"]
        url = f"{self.BASE_API_URL}/paper/{paper_id}/citations?{self._get_string_fields()}offset=1&limit={max_result}"  # noqa

        response = requests.get(url)
        if response.status_code == 200:
            res_data = response.json()["data"]
            final_result: List[Dict] = []

            for i, item in enumerate(res_data):
                if not item["citingPaper"].get("paperId"):
                    continue

                item["citingPaper"]["url"] = self._get_url(
                    item["citingPaper"]["paperId"]
                )
                item["citingPaper"]["id"] = item["citingPaper"]["paperId"]
                del item["citingPaper"]["paperId"]
                final_result.append(item["citingPaper"])

            logger.debug(f"[pne semantic scholar result] {json.dumps(final_result)}")
            return final_result
        raise NetWorkError("semantic scholar")



import sys
import warnings

from promptulate.tools.base import Tool
from promptulate.tools.shell.api_wrapper import ShellAPIWrapper


def _get_platform() -> str:
    """Get platform."""
    system = sys.platform
    if system == "Darwin":
        return "MacOS"
    return system


class ShellTool(Tool):
    """Tool to run shell commands."""

    name: str = "terminal"
    description: str = f"Run shell commands on this {_get_platform()} machine."

    def _run(self, command: str) -> str:
        warnings.warn(
            "The shell tool has no safeguards by default. Use at your own risk."
        )
        """Run commands and return final output."""
        return ShellAPIWrapper.run(command)



from enum import Enum
from typing import Any, Dict, List, Optional

from promptulate.pydantic_v1 import BaseModel, Field


class Status(Enum):
    created = "created"
    running = "running"
    completed = "completed"


class StepOutput(BaseModel):
    pass


class Artifact(BaseModel):
    artifact_id: str = Field(
        ...,
        description="ID of the artifact.",
        example="b225e278-8b4c-4f99-a696-8facf19f0e56",
    )
    agent_created: bool = Field(
        ...,
        description="Whether the artifact has been created by the agent.",
        example=False,
    )
    file_name: str = Field(
        ..., description="Filename of the artifact.", example="main.py"
    )
    relative_path: Optional[str] = Field(
        None,
        description="Relative path of the artifact in the agent's workspace.",
        example="python/code/",
    )


class StepRequestBody(BaseModel):
    input: Optional[str] = Field(
        None,
        description="Input prompt for the step.",
        example="Write the words you receive to the file 'output.txt'.",
    )
    additional_input: Optional[Dict[str, Any]] = None


class Step(StepRequestBody):
    task_id: str = Field(
        ...,
        description="The ID of the task this step belongs to.",
        example="50da533e-3904-4401-8a07-c49adf88b5eb",
    )
    step_id: str = Field(
        ...,
        description="The ID of the task step.",
        example="6bb1801a-fd80-45e8-899a-4dd723cc602e",
    )
    name: Optional[str] = Field(
        None, description="The name of the task step.", example="Write to file"
    )
    status: Status = Field(
        ..., description="The status of the task step.", example="created"
    )
    output: Optional[str] = Field(
        None,
        description="Output of the task step.",
        example="I am going to use the write_to_file command and write Washington to a file called output.txt <write_to_file('output.txt', 'Washington')",  # noqa
    )
    additional_properties: Optional[Dict[str, Any]] = None
    additional_output: Optional[StepOutput] = None
    artifacts: List[Artifact] = Field(
        [], description="A list of artifacts that the step has produced."
    )
    is_last: bool = Field(
        ..., description="Whether this is the last step in the task.", example=True
    )


class TaskRequestBody(BaseModel):
    input: Optional[str] = Field(
        None,
        description="Input prompt for the task.",
        example="Write 'Washington' to the file 'output.txt'.",
    )
    additional_input: Optional[Dict[str, Any]] = None


class Task(TaskRequestBody):
    task_id: str = Field(
        ...,
        description="The ID of the task.",
        example="50da533e-3904-4401-8a07-c49adf88b5eb",
    )
    steps: List[Step] = Field(..., description="A list of steps that the task has.")
    artifacts: List[Artifact] = Field(
        ...,
        description="A list of artifacts that the task has produced.",
        example=[
            "7a49f31c-f9c6-4346-a22c-e32bc5af4d8e",
            "ab7b4091-2560-4692-a4fe-d831ea3ca7d6",
        ],
    )



import uuid
from abc import ABC
from typing import Any, Dict, List, Optional, Union

from promptulate.uacp.schema import Artifact, Status, Step, Task
from promptulate.utils.logger import logger


class TaskDB(ABC):
    def create_task(
        self,
        input: Optional[str],
        additional_input: Any = None,
        artifacts: Optional[List[Artifact]] = None,
        steps: Optional[List[Step]] = None,
    ) -> Task:
        raise NotImplementedError

    def create_step(
        self,
        task_id: str,
        name: Optional[str] = None,
        input: Optional[str] = None,
        is_last: bool = False,
        additional_properties: Optional[Dict[str, Any]] = None,
        artifacts: Optional[List[Artifact]] = None,
    ) -> Step:
        raise NotImplementedError

    def create_artifact(
        self,
        task_id: str,
        file_name: str,
        agent_created: bool = True,
        relative_path: Optional[str] = None,
        step_id: Optional[str] = None,
    ) -> Artifact:
        raise NotImplementedError

    def get_task(self, task_id: str) -> Task:
        raise NotImplementedError

    def get_step(self, task_id: str, step_id: str) -> Step:
        raise NotImplementedError

    def get_artifact(self, task_id: str, artifact_id: str) -> Artifact:
        raise NotImplementedError

    def update_task(self, task: Task) -> Task:
        raise NotImplementedError

    def update_step(self, task_id: str, step: Step) -> Step:
        raise NotImplementedError

    def list_tasks(self) -> List[Task]:
        raise NotImplementedError

    def list_steps(self, task_id: str, status: Optional[Status] = None) -> List[Step]:
        raise NotImplementedError


class InMemoryTaskDB(TaskDB):
    _tasks: Dict[str, Task] = {}

    def create_task(
        self,
        input: Optional[str],
        additional_input: Any = None,
        artifacts: Optional[List[Artifact]] = None,
        steps: Optional[List[Step]] = None,
    ) -> Task:
        steps = steps or []
        artifacts = artifacts or []
        task_id = str(uuid.uuid4())

        task = Task(
            task_id=task_id,
            input=input,
            steps=steps,
            artifacts=artifacts,
            additional_input=additional_input,
        )
        self._tasks[task_id] = task
        return task

    def get_task(self, task_id: str) -> Task:
        task = self._tasks.get(task_id, None)
        if not task:
            raise ValueError(f"Task {task_id} not found")
        return task

    def update_task(self, task: Task) -> Task:
        if self._tasks.get(task.task_id, None) is None:
            raise ValueError(f"Task {task.task_id} not found")
        self._tasks[task.task_id] = task
        return task

    def create_step(
        self,
        task_id: str,
        name: Optional[str] = None,
        input: Optional[str] = None,
        is_last=False,
        additional_properties: Optional[Dict[str, Any]] = None,
        artifacts: Optional[List[Artifact]] = None,
    ) -> Step:
        step_id: str = str(uuid.uuid4())
        artifacts: List[Artifact] = artifacts or []

        step = Step(
            task_id=task_id,
            step_id=step_id,
            name=name,
            input=input,
            status=Status.created,
            is_last=is_last,
            additional_properties=additional_properties,
            artifacts=artifacts,
        )
        logger.info(f"Create step: {step.json()}")
        task = self.get_task(task_id)
        task.steps.append(step)
        return step

    def get_step(self, task_id: str, step_id: str) -> Step:
        task = self.get_task(task_id)
        step = next(filter(lambda s: s.task_id == task_id, task.steps), None)
        if not step:
            raise ValueError(f"Step {step_id} not found in task {task_id}")
        return step

    def update_step(self, task_id: str, step: Step) -> Step:
        task = self.get_task(task_id)

        for i, s in enumerate(task.steps):
            if s.step_id == step.step_id:
                task.steps[i] = step
                return step

        raise ValueError(f"Step {step.step_id} not found in task {task_id}")

    def create_artifact(
        self,
        task_id: str,
        file_name: str,
        agent_created: bool = True,
        relative_path: Optional[str] = None,
        step_id: Optional[str] = None,
    ) -> Artifact:
        artifact_id = str(uuid.uuid4())
        artifact = Artifact(
            artifact_id=artifact_id,
            agent_created=agent_created,
            file_name=file_name,
            relative_path=relative_path,
        )
        task = self.get_task(task_id)
        task.artifacts.append(artifact)

        if step_id:
            step = self.get_step(task_id, step_id)
            step.artifacts.append(artifact)

        return artifact

    def get_artifact(self, task_id: str, artifact_id: str) -> Artifact:
        task = self.get_task(task_id)
        artifact = next(
            filter(lambda a: a.artifact_id == artifact_id, task.artifacts), None
        )
        if not artifact:
            raise ValueError(f"Artifact {artifact_id} not found in task {task_id}")
        return artifact

    def list_tasks(self) -> List[Task]:
        return [task for task in self._tasks.values()]

    def list_steps(self, task_id: str, status: Optional[Status] = None) -> List[Step]:
        task = self.get_task(task_id)
        steps = task.steps
        if status:
            steps = list(filter(lambda s: s.status == status, steps))
        return steps



# Copyright (c) 2023 promptulate
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# Copyright Owner: Zeeland
# GitHub Link: https://github.com/Undertone0809/
# Project Link: https://github.com/Undertone0809/promptulate
# Contact Email: zeeland@foxmail.com

from typing import List, Optional

from promptulate.frameworks.prompt import (
    SUMMARY_CONTENT_PROMPT_ZH,
    SUMMARY_TOPIC_PROMPT_ZH,
)
from promptulate.provider.base import BaseMixin
from promptulate.pydantic_v1 import BaseModel
from promptulate.schema import BaseMessage, MessageSet, UserMessage


class SummarizerMixin(BaseMixin):
    """message summary capability provider"""

    def summary_content(
        self,
        enable_embed_message: bool = False,
        summary_prompt: str = SUMMARY_CONTENT_PROMPT_ZH,
    ):
        message_history: MessageSet = self.memory.load_message_set_from_memory()
        message_history.messages.append(UserMessage(content=summary_prompt))
        assistant_answer: BaseMessage = self.llm.predict(message_history)
        if enable_embed_message:
            self.embed_message(assistant_answer, message_history)
        return assistant_answer.content

    def summary_topic(
        self,
        enable_embed_message: bool = False,
        summary_topic_prompt: str = SUMMARY_TOPIC_PROMPT_ZH,
    ):
        message_history: MessageSet = self.memory.load_message_set_from_memory()
        message_history.messages.append(UserMessage(content=summary_topic_prompt))
        assistant_answer: BaseMessage = self.llm.predict(message_history)
        if enable_embed_message:
            self.embed_message(assistant_answer, message_history)
        return assistant_answer.content


class TranslatorMixin(BaseMixin):
    """let the llm answer question in the specified language"""

    def predict_by_translate(
        self, prompt: str, country: str, enable_embed_message: bool = False
    ):
        """
        predict by specified language

        Args:
            enable_embed_message: Whether to save this session in the history session
            prompt: you prompt
            country: which country's language you want to export

        Returns:
            the country official language you choose
        """
        message_history: MessageSet = self.memory.load_message_set_from_memory()
        message_history.messages.append(
            UserMessage(
                content=f"{prompt}. Please answer question using {country} official language. "  # noqa: E501
            )
        )
        assistant_answer: BaseMessage = self.llm.predict(message_history)
        if enable_embed_message:
            self.embed_message(assistant_answer, message_history)
        return assistant_answer.content


class DeriveHistoryMessageMixin(BaseMixin):
    """provide history message output as markdown"""

    def get_history(self) -> List[dict]:
        """get history conversation from memory"""
        return self.memory.load_message_set_from_memory().listdict_messages

    def export_message_to_markdown(
        self, output_type: str = "text", file_path: str = "output.md"
    ) -> Optional[str]:
        """
        convert message to the string type or file type markdown

        Args:
            output_type: text or file. default is text
            file_path: output file path

        Returns:
            string type conversation in markdown format
        """
        message_history: List[dict] = self.get_history()

        ret = "# Chat record\n"
        for message in message_history:
            role = message.get("preset_roles")
            content = message.get("content").replace('"', '\\"')
            if role == "assistant":
                ret += f"## Bot said\n\n{content}\n\n"
            else:
                ret += f"## You said\n\n{content}\n\n"
        if output_type == "text":
            return ret
        elif output_type == "file":
            with open(file_path, "w", encoding="utf-8") as f:
                f.write(ret)
        else:
            raise ValueError(
                "Invalid output destination specified. Please choose either 'text' or 'file'."  # noqa: E501
            )
        return ret


class StorageHistoryMessageMixin(BaseModel):
    pass



from abc import abstractmethod
from enum import Enum
from typing import Any, Callable, Dict, Iterator, List, Optional, Union

from promptulate.pydantic_v1 import BaseModel, Field

__all__ = [
    "LLMType",
    "BaseMessage",
    "CompletionMessage",
    "SystemMessage",
    "UserMessage",
    "AssistantMessage",
    "MessageSet",
    "init_chat_message_history",
    "StreamIterator",
]


class BaseMessage(BaseModel):
    """Message basic object."""

    content: str
    additional_kwargs: dict = Field(default_factory=dict)

    @property
    @abstractmethod
    def type(self) -> str:
        """Type of the message, used for serialization."""


class StreamIterator:
    """
    This class is an iterator for the response stream from the LLM model.

    Attributes:
        response_stream: The stream of responses from the LLM model.
        parse_content: The callback function to parse the chunk.
        return_raw_response: A boolean indicating whether to return the raw response
        or not.
        additional_kwargs: Optional dictionary with additional keyword parameters
        content: An optional string that represents the content
    """

    def __init__(
        self,
        response_stream,
        parse_content: callable([[Any], [str, str]]),
        return_raw_response: bool = False,
        additional_kwargs: dict = None,
        content: str = None,
    ):
        """
        The constructor for BaseStreamIterator class.

        Args:
            response_stream: The stream of responses from the LLM model.
            parse_content: The callback function to parse the chunk.
            return_raw_response: A boolean indicating whether to return the raw response
            or not.
            additional_kwargs: Optional dictionary with additional keyword parameters
            content: An optional string that represents the content
        """
        self.response_stream = response_stream
        self.return_raw_response = return_raw_response
        self.parse_content = parse_content
        self.additional_kwargs = additional_kwargs or {}
        self.content = content

    def __iter__(self) -> Union[Iterator[BaseMessage], Iterator[str]]:
        """
        The iterator method for the BaseStreamIterator class.

        Returns:
            self: An instance of the BaseStreamIterator class.
        """
        return self

    def parse_chunk(self, chunk) -> Optional[Union[str, BaseMessage]]:
        """
        This method is used to parse a chunk from the response stream. It returns
        None if the chunk is empty, otherwise it returns the parsed chunk.

        Args:
            chunk: The chunk to be parsed.

        Returns:
            Optional: The parsed chunk or None if the chunk is empty.
        """
        content, ret_data = self.parse_content(chunk)
        if content is None:
            return None
        if self.return_raw_response:
            additional_kwargs: dict = ret_data
            message = AssistantMessage(
                content=content,
                additional_kwargs=additional_kwargs,
            )
            return message

        return content

    def __next__(self) -> Union[str, BaseMessage]:
        """
        The next method for the BaseStreamIterator class.

        This method is used to get the next response from the LLM model. It iterates
        over the response stream and parses each chunk using the parse_chunk method.
        If the parsed chunk is not None, it returns the parsed chunk as the next
        response. If there are no more messages in the response stream, it raises a
        StopIteration exception.

        Returns:
            Union[str, BaseMessage]: The next response from the LLM model. If
            return_raw_response is True, it returns an AssistantMessage instance,
            otherwise it returns the content of the response as a string.
        """
        for chunk in self.response_stream:
            content, ret_data = self.parse_content(chunk)
            if content is None:
                continue
            if self.return_raw_response:
                additional_kwargs: dict = ret_data
                message = AssistantMessage(
                    content=content,
                    additional_kwargs=additional_kwargs,
                )
                return message

            return content

        # If there are no more messages, stop the iteration
        raise StopIteration


class CompletionMessage(BaseMessage):
    """Type of completion message. Used in OpenAI currently"""

    @property
    def type(self) -> str:
        return "completion"


class SystemMessage(BaseMessage):
    """Type of message that is a system message. Currently used in OpenAI."""

    @property
    def type(self) -> str:
        """Type of the message, used for serialization."""
        return "system"


class UserMessage(BaseMessage):
    """Type of message that is a user message. Currently used in OpenAI."""

    @property
    def type(self) -> str:
        return "user"


class AssistantMessage(BaseMessage):
    """Type of message that is an assistant message. Currently used in OpenAI."""

    @property
    def type(self) -> str:
        return "assistant"


MESSAGE_TYPE = {
    "completion": CompletionMessage,
    "system": SystemMessage,
    "user": UserMessage,
    "assistant": AssistantMessage,
}


class LLMType(str, Enum):
    """All LLM type here"""

    OpenAI = "OpenAI"
    ChatOpenAI = "ChatOpenAI"
    ErnieBot = "ErnieBot"
    QianFan = "QianFan"
    ZhiPu = "ZhiPu"


class MessageSet:
    """MessageSet can be used in Memory, LLMs, Framework and some else.
    It's a universal chat message format in promptulate.
    """

    def __init__(
        self,
        messages: List[BaseMessage],
        conversation_id: Optional[str] = None,
        additional_kwargs: Optional[dict] = None,
    ):
        self.messages: List[BaseMessage] = messages
        self.conversation_id: Optional[str] = conversation_id
        self.additional_kwargs: dict = additional_kwargs or {}

    @classmethod
    def from_listdict_data(
        cls, value: List[Dict], additional_kwargs: Optional[dict] = None
    ) -> "MessageSet":
        """initialize MessageSet from a List[Dict] data

        Args:
            value(List[Dict]): the example is as follows:
                [
                    {"type": "user", "content": "This is a message1."},
                    {"type": "assistant", "content": "This is a message2."}
                ]
            additional_kwargs(Optional[dict]): additional kwargs

        Returns:
            initialized MessageSet
        """
        messages: List[BaseMessage] = [
            MESSAGE_TYPE[item["role"]](content=item["content"]) for item in value
        ]
        return cls(messages=messages, additional_kwargs=additional_kwargs)

    @property
    def listdict_messages(self) -> List[Dict[str, str]]:
        """Convert the MessageSet messages to a list of dictionary(openai type).

        Returns:
            List[Dict[str, str]]: the example is as follows:
                [
                    {"role": "user", "content": "This is a message1."},
                    {"role": "assistant", "content": "This is a message2."}
                ]
        """
        converted_messages = []
        for message in self.messages:
            converted_messages.append(
                {"role": message.type, "content": message.content}
            )
        return converted_messages

    def to_llm_prompt(self, llm_type: LLMType) -> Any:
        """Convert the MessageSet messages to specified llm prompt"""
        if not llm_type:
            ValueError(
                "Missing llm_type, llm_type is needed if you want to use llm_prompt."
            )
        return _to_llm_prompt[llm_type](self)

    @property
    def string_messages(self) -> str:
        """Convert the message to a string type, it can be used as a prompt for OpenAI
        completion."""
        string_result = ""
        for message in self.messages:
            string_result += f"{message.content}\n"
        return string_result

    def add_message(self, message: BaseMessage) -> None:
        self.messages.append(message)

    def add_completion_message(self, message: str) -> None:
        self.messages.append(CompletionMessage(content=message))

    def add_system_message(self, message: str) -> None:
        self.messages.append(SystemMessage(content=message))

    def add_user_message(self, message: str) -> None:
        self.messages.append(UserMessage(content=message))

    def add_ai_message(self, message: str) -> None:
        self.messages.append(AssistantMessage(content=message))


def init_chat_message_history(
    system_content: str, user_content: str, llm: LLMType
) -> MessageSet:
    if llm == llm.ChatOpenAI or llm == llm.OpenAI:
        messages = [
            SystemMessage(content=system_content),
            UserMessage(content=user_content),
        ]
    else:
        messages = [
            UserMessage(content=system_content),
            AssistantMessage(content="好的"),
            UserMessage(content=user_content),
        ]
    return MessageSet(messages=messages)


def _to_openai_llm_prompt(message_set: MessageSet) -> str:
    return message_set.string_messages


def _to_chat_openai_llm_prompt(message_set: MessageSet) -> List[Dict]:
    return message_set.listdict_messages


def _to_ernie_bot_llm_prompt(message_set: MessageSet) -> List[Dict]:
    return message_set.listdict_messages


def _to_qian_fan_llm_prompt(message_set: MessageSet) -> List[Dict]:
    return message_set.listdict_messages


def _to_zhipu_llm_prompt(message_set: MessageSet) -> List[Dict]:
    return message_set.listdict_messages


_to_llm_prompt: Dict[LLMType, Callable] = {
    LLMType.OpenAI: _to_openai_llm_prompt,
    LLMType.ChatOpenAI: _to_chat_openai_llm_prompt,
    LLMType.ErnieBot: _to_ernie_bot_llm_prompt,
    LLMType.QianFan: _to_qian_fan_llm_prompt,
    LLMType.ZhiPu: _to_zhipu_llm_prompt,
}



from promptulate.utils.string_template import StringTemplate

SYSTEM_PROMPT_TEMPLATE = StringTemplate(
    template_format="jinja2",
    template="""As a diligent Task Agent, you goal is to effectively accomplish the provided task or question as best as you can.

## Tools
You have access to the following tools, the tools information is provided by the following schema:
{{tool_descriptions}}

## Task
Currently, you are working on the following task:
{{task}}

To achieve your goals, you need to choose the appropriate tools for reasoning.
For example: If the user wants to check the weather in Beijing tomorrow. The first step is to use websearch to query the weather in Beijing. After obtaining the results, in the second step, you can use the finish command to return the results.

## Constraints
- Choose only ONE tool in one step.
- Choose tool carefully as it is critical to accomplish the task.
- Your final answer output language should be consistent with the language used by the user. Middle step output is English.

{{current_process}}

{{output_format}}
""",  # noqa: E501
)

REACT_SYSTEM_PROMPT_TEMPLATE = StringTemplate(
    template_format="jinja2",
    template="""
As a diligent Task Agent, you goal is to effectively accomplish the provided task or question as best as you can.

## Tools
You have access to the following tools, the tools information is provided by the following schema:
{{tool_descriptions}}

## Output Format
To answer the question, Use the following JSON format. JSON only, no explanation. Otherwise, you will be punished.
The output should be formatted as a JSON instance that conforms to the format below. JSON only, no explanation.

```json
{
"analysis": "The thought of what to do and why.",
"action": # the action to take, must be one of provided tools
    {
    "name": "tool name",
    "args": "tool input parameters, json type data"
    }
}
```

If this format is used, the user will respond in the following format:

```
Observation: tool response
```

You should keep repeating the above format until you have enough information
to answer the question without using any more tools. At that point, you MUST respond
in the one of the following two formats:

- If you can answer the question:

```json
{
"analysis": "The thought of what to do and why.",
"action": {
    "name": "finish",
    "args": {"content": "You answer here."}
    }
}
```

- If you cannot answer the question in the current context:

```json
{
"thought": "The thought of what to do and why.",
"action": {
    "name": "finish",
    "args": {"content": "Sorry, I cannot answer your query, because (Summary all the upper steps, and explain)"}
    }
}
```

## Attention
- Your output is JSON only and no explanation.
- Choose only ONE tool and you can't do without using any tools in one step.
- Your final answer output language should be consistent with the language used by the user. Middle step output is English.
- Whether the action input is JSON or str depends on the definition of the tool.

## User question
{{question}}

## Current Conversation
Below is the current conversation consisting of interleaving human and assistant history.
""",  # noqa: E501
)
PREFIX_TEMPLATE = """You are a {agent_identity}, named {agent_name}, your goal is {agent_goal}, and the constraint is {agent_constraints}. """  # noqa



PLAN_SYS_PROMPT = """
For the given objective, come up with a simple step by step plan.
This plan should involve individual tasks, that if executed correctly will yield the correct answer. Do not add any superfluous steps.
The result of the final step should be the final answer. Make sure that each step has all the information needed - do not skip steps.
"""  # noqa



from typing import TYPE_CHECKING, Any, List, Optional

from promptulate.pydantic_v1 import BaseModel
from promptulate.tools.base import Tool

if TYPE_CHECKING:
    from langchain.tools.base import BaseTool as LangchainBaseToolType


class LangchainTool(Tool):
    def __init__(self, langchain_tool: "LangchainBaseToolType", **kwargs):
        """Wrap langchain tool to promptulate tool.

        Args:
            langchain_tool(LangchainBaseToolType): Langchain tool.
        """
        from langchain.tools.base import BaseTool as LangchainBaseTool

        if not isinstance(langchain_tool, LangchainBaseTool):
            raise ValueError("langchain_tool should be type of langchain tool.")

        self.langchain_tool: "LangchainBaseToolType" = langchain_tool
        self.name: str = self.langchain_tool.name
        self.description: str = self.langchain_tool.description
        self.parameters: Optional[BaseModel] = self.langchain_tool.args_schema
        super().__init__(**kwargs)

    def _run(self, *args, **kwargs) -> Any:
        """Run the langchain tool.

        Args:
            *args: if the first arg is str, it will be used as tool_input.
            **kwargs: if the first arg is not str, it will be used as tool_input.

        Returns:
            Any: The result of the langchain tool.
        """
        if args and isinstance(args[0], str):
            return self.langchain_tool.run(tool_input=args[0])

        return self.langchain_tool.run(tool_input=kwargs)



# Copyright (c) 2023 promptulate
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# Copyright Owner: Zeeland
# GitHub Link: https://github.com/Undertone0809/
# Project Link: https://github.com/Undertone0809/promptulate
# Contact Email: zeeland@foxmail.com

__all__ = [
    "SUMMARY_CONTENT_PROMPT_ZH",
    "SUMMARY_TOPIC_PROMPT_ZH",
    "SUMMARY_TOPIC_PROMPT_EN",
    "SUMMARY_CONTENT_PROMPT_EN",
]

SUMMARY_CONTENT_PROMPT_ZH = """
简要总结一下你和用户的对话，用作后续的上下文提示 prompt，控制在 200 字以内
"""

SUMMARY_TOPIC_PROMPT_ZH = """
上面是ai 和用户的历史聊天总结作为前情提要，请使用四到五个字直接返回这句话的简要主题，
不要解释、不要标点、不要语气词、不要多余文本，如果没有主题，请直接返回“闲聊”
"""

SUMMARY_CONTENT_PROMPT_EN = """
Give a quick summary of your conversation with the user and use it as a follow-up context prompt, no more than 200 words
"""  # noqa

SUMMARY_TOPIC_PROMPT_EN = """
As the previous feed, please use four or five words to return directly to the brief topic of the sentence,
no explanation, no punctuation, no particles, no extra text, or if there is no topic, just return to "small talk".
"""  # noqa



from promptulate.utils.string_template import StringTemplate

SYSTEM_PROMPT = """
Answer the following questions as best you can. You have access use web search.
After the user enters a question, you need to generate keywords for web search,
and then summarize until you think you can answer the user's answer.

Your output format is as follows:
Question: the input question you must answer
Thought: The next you should do
Query: web search query words
Observation: the result of query
... (this Thought/Query/Observation can repeat N times)
Thought: I know the final answer
Final Answer: the final answer to the original input question

Begin!

Question: {prompt}

Thought:"""

SYSTEM_PROMPT_TEMPLATE = StringTemplate(SYSTEM_PROMPT)



import json
import math
import re

import numexpr

from promptulate.llms.base import BaseLLM
from promptulate.llms.openai import ChatOpenAI
from promptulate.tools.base import Tool
from promptulate.tools.math.prompt import prompt_template


def _evaluate_expression(expression: str) -> str:
    """
    Parse numexpr expression.

    This function takes a string expression, evaluates it using `numexpr.evaluate`,
    and returns the result as a string. It also handles exceptions and raises a
    ValueError with a custom error message if the evaluation fails.

    Args:
        expression (str): The expression to evaluate.

    Returns:
        str: The result of the evaluation.

    Raises:
        ValueError: If the evaluation fails.
    """
    try:
        local_dict = {"pi": math.pi, "e": math.e}
        output = str(
            numexpr.evaluate(
                expression.strip(),
                global_dict={},  # restrict access to globals
                local_dict=local_dict,  # add common mathematical functions
            )
        )
    except Exception as e:
        raise ValueError(
            f'Calculator._evaluate_expression("{expression}") raised error: {e}.'
            " Please try again with a valid numerical expression"
        )

    # Remove any leading and trailing brackets from the output
    return re.sub(r"^\[|\]$", "", output)


def _is_valid_expression(expression: str) -> bool:
    """Check if the expression is valid."""
    try:
        _evaluate_expression(expression)
        return True
    except ValueError:
        return False


class Calculator(Tool):
    """
    A Math operator.

    This class is a tool for evaluating mathematical expressions. It uses the
    _evaluate_expression and _is_valid_expression functions to evaluate expressions
    and check their validity. It also uses the BaseLLM class to generate prompts
    for the user.

    Attributes:
        name (str): The name of the tool.
        description (str): A description of the tool.
        llm (BaseLLM): An instance of the BaseLLM class.
    """

    name: str = "math-calculator"
    description: str = (
        "Useful for when you need to answer questions about math.You input is a nature"
        "language of math expression. Attention: Expressions can not exist variables!"
        "eg: (current age)^0.43 is wrong, you should use 18^0.43 instead."
    )

    def __init__(self, llm: BaseLLM = None, **kwargs):
        """
        Initialize the Calculator class.

        This method initializes the Calculator class with an instance of the BaseLLM
        class and any additional keyword arguments.

        Args:
            llm (BaseLLM, optional): An instance of the BaseLLM class. Defaults to None.
        """
        self.llm: BaseLLM = llm or ChatOpenAI(
            temperature=0, enable_default_system_prompt=False
        )
        super().__init__(**kwargs)

    def _run(self, question: str) -> str:
        """
        Run the Calculator tool.

        This method takes a prompt from the user, checks if it is a valid expression,
        and evaluates it if it is. If it is not a valid expression, it generates a
        new prompt using the BaseLLM class and evaluates the resulting expression.

        Args:
            question (str): The math problem of user.

        Returns:
            str: The result of the evaluation.

        Raises:
            ValueError: If the evaluation fails.
        """
        if _is_valid_expression(question):
            return _evaluate_expression(question)

        prompt: str = prompt_template.format(question=question)
        llm_output: str = self.llm(prompt)
        try:
            expression = json.loads(llm_output).get("expression")
            return _evaluate_expression(expression)
        except Exception as e:
            raise ValueError(f"Unknown format from LLM: {llm_output}, error: {e}")


def calculator(expression: str):
    """Evaluate a mathematical expression.

    This function takes a string expression, evaluates it using `numexpr.evaluate`,
    and returns the result as a string. It also handles exceptions and raises a
    ValueError with a custom error message if the evaluation fails.

    Args:
        expression: A mathematical expression, eg: 18^0.43

    Attention:
        Expressions can not exist variables!
        bad example: (current age)^0.43, (number)^(1/5)
        good example: 18^0.43, 37593**(1/5)

    Returns:
        The result of the evaluation.
    """
    return _evaluate_expression(expression)



import re
from time import sleep


def sleep_tool(sleep_time: str) -> str:
    """Make agent sleep for a specified number of seconds.
       Input is a number. eg: Sleep for 5s and enter 5
    Args:
        sleep_time: Sleep for sleep_time

    Returns:
        str: Agent slept seconds.
    """
    sleep_time = re.findall(r"\d+", sleep_time)
    sleep(int(sleep_time[0]))
    return f"Agent slept for {sleep_time[0]} seconds."



# Copyright (c) 2023 promptulate
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# Copyright Owner: Zeeland
# GitHub Link: https://github.com/Undertone0809/
# Project Link: https://github.com/Undertone0809/promptulate
# Contact Email: zeeland@foxmail.com

from typing import Optional

from promptulate.config import Config

PROXY_MODE = ["off", "custom", "promptulate"]


def set_proxy_mode(mode: str, proxies: Optional[dict] = None):
    """
    Set proxy mode. Promptulate offer three proxy mode ['off', 'custom', 'promptulate']
    'off' disables your proxy mode. This is the normal status.
    'custom' means you can set your custom proxy. Moreover, you must pass proxies
    'promptulate' provide free proxy you can use it.

    Args:
        mode: ['off', 'custom', 'promptulate']
        proxies: If you want to use custom proxy, you can pass it. An example
        proxies is {'http': 'http://127.0.0.1:7890'}
    """
    if mode not in PROXY_MODE:
        raise ValueError("proxy mode must in ['off', 'custom', 'promptulate']")

    if mode == "promptulate":
        raise ValueError(
            "Current promptulate free proxy mode is not supported yet. A better solution will be provided soon."  # noqa
        )

    if mode == "custom" and not proxies:
        raise ValueError("proxies must be passed when proxy mode is custom")

    Config().set_proxy_mode(mode, proxies)



from typing import List

from promptulate.tools.arxiv.tools import (
    ArxivQueryTool,
    ArxivReferenceTool,
    ArxivSummaryTool,
)
from promptulate.tools.base import BaseToolKit, Tool


class ArxivTootKit(BaseToolKit):
    def get_tools(self) -> List[Tool]:
        return [
            ArxivSummaryTool(),
            ArxivQueryTool(),
            ArxivReferenceTool(),
        ]



from typing import Dict, List, Optional, Union

import arxiv

from promptulate.pydantic_v1 import BaseModel, root_validator


class ArxivQuerySet:
    def __init__(self, search: arxiv.Search, keys: Optional[List[str]] = None):
        self.search: arxiv.Search = search
        self._data: List[Dict] = []

        for result in search.results():
            if keys:
                filtered_result_dict = {
                    k: result.__dict__[k] for k in keys if k in result.__dict__
                }
                self._data.append(filtered_result_dict)
            else:
                self._data.append(result.__dict__)

    def __str__(self):
        result: str = ""
        for item in self._data:
            result += f"{item.__dict__}\n"
        return result

    def titles(self) -> List[str]:
        titles: List[str] = []
        for item in self.search.results():
            titles.append(item.title)
        return titles

    def all(self) -> List[Dict]:
        return self._data

    def first(self) -> Optional[Dict]:
        if len(self._data) == 0:
            return None
        return self._data[0]

    @classmethod
    def from_filter_result(cls, search: arxiv.Search, *args):
        """generate a filter key queryset"""
        return cls(search, list(args))


class ArxivAPIWrapper(BaseModel):
    """https://github.com/lukasschwab/arxiv.py"""

    max_num_of_result: int = 5
    sort_by = arxiv.SortCriterion.Relevance

    class Config:
        """Configuration for this pydantic object."""

        arbitrary_types_allowed = True

    @root_validator(skip_on_failure=True)
    def validate_environment(cls, values: Dict) -> Dict:
        """Validate that python package exists in environment."""
        try:
            import arxiv  # noqa
        except ImportError:
            raise ValueError(
                "Could not import arxiv python package. "
                "Please install it with `pip install arxiv`."
            )
        return values

    def _query(
        self,
        keyword: Optional[str] = None,
        id_list: Union[List[str], str, None] = None,
        num_results: Optional[int] = None,
    ) -> arxiv.Search:
        """
        query arxiv paper by keyword or id_list
        Args:
            keyword: query keyword
            id_list: arxiv id, you can input multiple id or single
        Returns:
            arxiv search type data result
        """
        if not keyword:
            keyword = ""

        if not id_list:
            id_list = []

        if not num_results:
            num_results = self.max_num_of_result

        if isinstance(id_list, str):
            id_list = [id_list]

        search = arxiv.Search(
            query=keyword,
            id_list=id_list,
            max_results=num_results,
            sort_by=self.sort_by,
        )
        return search

    def query(
        self,
        keyword: Optional[str] = None,
        id_list: Union[List[str], str, None] = None,
        num_results: Optional[int] = None,
        **kwargs,
    ) -> List[Dict]:
        """
        Query arxiv paper by keyword or id_list. You can make ArxivQuerySet return the
        specified fields from arxiv query result.
        Args:
            num_results: max num of result. Return self.max_num_of_result if none
            keyword: query keyword
            id_list: arxiv id, you can input multiple id or single
            kwargs:
                specified_fields(Optional[List[str]]): filter specified field to return.
                For example, you can return the ["title", "summary"] fields from each
                arxiv query result.
        Returns:
            List[Dict] type result
        Examples:
            If you want to get paper title and summary, you can do as follows.

            >> arxiv_api_wrapper = ArxivAPIWrapper()
            >> result = arxiv_api_wrapper.query("LLM", specified_fields=["entry_id", "title", "summary"])

            All fields you can see in: https://github.com/lukasschwab/arxiv.py
            Common fields are: ["entry_id", "title", "authors", "summary", "published"]
        """  # noqa
        search = self._query(keyword, id_list, num_results)
        if "specified_fields" in kwargs:
            return ArxivQuerySet.from_filter_result(
                search, *kwargs["specified_fields"]
            ).all()
        if "from_callback" in kwargs and kwargs["from_callback"] == "arxiv-query":
            keys = ["entry_id", "title", "authors", "summary"]
            return ArxivQuerySet.from_filter_result(search, *keys).all()
        return ArxivQuerySet(search).all()

    def download_pdf(self, id_list: List[str]) -> str:
        paper = next(arxiv.Search(id_list=id_list).results())
        return paper.download_pdf()



import os
import shutil
from typing import Optional

from promptulate.pydantic_v1 import BaseModel, Field
from promptulate.tools.base import Tool


class RToolParameters(BaseModel):
    file_name: str = Field(
        description="The name of the file you want to edit.",
    )


class ToolParameters(RToolParameters):
    text: Optional[str] = Field(
        default=None,
        description="The content you want to edit.",
    )


class CMToolParameters(RToolParameters):
    destination_path: str = Field(
        description="The directory you want to reach.",
    )


class WriteFileTool(Tool):
    name: str = "write-file"
    description: str = (
        "Write to a file"
        "Useful when you need to edit/create and edit a file."
        "It can edit the specified file in the specified directory/local directory"
        "If the file does not exist, edit it after it is created."
    )
    parameters = ToolParameters

    def __init__(self, root_dir: str, *args, **kwargs) -> None:
        self.root_dir = root_dir
        super().__init__(*args, **kwargs)

    def _run(self, file_name: str, text: Optional[str] = None) -> str:
        """Write to a file

        Args:
            file_name(Optional(str)): The name of the file you want to edit.
            text(Optional(str)): The content you want to edit.

        Returns:
            str: Write file successfully
        """
        try:
            with open(f"{self.root_dir}/{file_name}", "w") as file:
                file.write(text)
        except IOError:
            return "An error occurred while trying to write to the file."
        except PermissionError:
            return "Permission denied: You don't have permission to write to this file."
        return "Write file successfully"


class AppendFileTool(Tool):
    name: str = "append-file"
    description: str = (
        "Append to a file"
        "Useful when you need to edit/create edit a file."
        "It can edit the specified file in the specified directory/local directory,"
        "Append the content to the end of the file."
        "If the file does not exist, edit it after it is created"
    )
    parameters = ToolParameters

    def __init__(self, root_dir: str, *args, **kwargs) -> None:
        self.root_dir = root_dir
        super().__init__(*args, **kwargs)

    def _run(self, file_name: str, text: Optional[str] = None) -> str:
        """Append to a file

        Args:
            file_name(Optional(str)): The name of the file you want to edit.
            text(Optional(str)): The content you want to edit.

        Returns:
            str: Append file successfully
        """
        try:
            with open(f"{self.root_dir}/{file_name}", "a") as file:
                file.write(text)
        except IOError:
            return "An error occurred while trying to write to the file."
        except PermissionError:
            return "Permission denied: You don't have permission to write to this file."
        return "Append file successfully"


class ReadFileTool(Tool):
    name: str = "read-file"
    description: str = (
        "Read a file"
        "Useful when you need to read a file."
        "It can read the specified file in the specified directory/local directory."
    )
    parameters = RToolParameters

    def __init__(self, root_dir: str, *args, **kwargs) -> None:
        self.root_dir = root_dir
        super().__init__(*args, **kwargs)

    def _run(self, file_name: str) -> str:
        """Read a file

        Args:
            file_name(Optional(str)): The name of the file you want to read.

        Returns:
            str: The contents of the file.
        """
        try:
            with open(f"{self.root_dir}/{file_name}", "r") as file:
                return file.read()
        except FileNotFoundError:
            return "File not found: The file you are trying to read does not exist."
        except PermissionError:
            return "Permission denied: You don't have permission to read this file."
        except UnicodeDecodeError:
            return (
                "Unicode decode error:"
                "The file could not be decoded with utf-8 encoding."
            )
        except IOError:
            return "An error occurred while trying to read the file."


class DeleteFileTool(Tool):
    name: str = "delete-file"
    description: str = (
        "Delete a file"
        "Useful when you need to delete a file."
        "It can delete the specified file in the specified directory/local directory."
    )
    parameters = RToolParameters

    def __init__(self, root_dir: str, *args, **kwargs) -> None:
        self.root_dir = root_dir
        super().__init__(*args, **kwargs)

    def _run(self, file_name: str) -> str:
        """Delete a file

        Args:
            file_name(Optional(str)): The name of the file you want to delete.

        Returns:
            str: Delete file successfully
        """
        try:
            os.remove(f"{self.root_dir}/{file_name}")
        except FileNotFoundError:
            return "File not found: The file you are trying to read does not exist."
        except PermissionError:
            return "Permission denied: You don't have permission to read this file."
        except IOError:
            return "An error occurred while trying to delete the file."
        return "Delete file successfully"


class ListDirectoryTool(Tool):
    name: str = "list-directory"
    description: str = (
        "List directory"
        "Useful when you need to list all files in the directory."
        "It can list all files in the specified directory/local directory."
    )
    parameters = None

    def __init__(self, root_dir: str, *args, **kwargs) -> None:
        self.root_dir = root_dir
        super().__init__(*args, **kwargs)

    def _run(self) -> str:
        """List directory

        Args:
            None

        Returns:
            str: All files in the current path(Separated by newlines)
        """
        try:
            entries = os.listdir(self.root_dir)
        except PermissionError:
            return "Permission denied: You don't have permission to read this file."
        return "\n".join(entries)


class CopyFileTool(Tool):
    name: str = "copy-file"
    description: str = (
        "Copy a file"
        "Useful when you need to copy a file."
        "It can copy the specified file in the specified directory/local directory."
    )
    parameters = CMToolParameters

    def __init__(self, root_dir: str, *args, **kwargs) -> None:
        self.root_dir = root_dir
        super().__init__(*args, **kwargs)

    def _run(self, file_name: str, destination_path: str) -> str:
        """Copy a file

        Args:
            file_name(Optional(str)): The name of the file you want to copy.
            destination_path(Optional(str)): The path you want to copy the file to.

        Returns:
            str: Copy file successfully
        """
        try:
            shutil.copy(
                f"{self.root_dir}/{file_name}", f"{destination_path}/{file_name}"
            )
        except FileNotFoundError:
            return "File not found: The file you are trying to read does not exist."
        except PermissionError:
            return "Permission denied: You don't have permission to read this file."
        except IOError:
            return "An error occurred while trying to copy the file."
        return "Copy file successfully"


class MoveFileTool(Tool):
    name: str = "move-file"
    description: str = (
        "Move a file"
        "Useful when you need to move a file."
        "It can move or rename the specified file \
        in the specified directory/local directory."
    )
    parameters = CMToolParameters

    def __init__(self, root_dir: str, *args, **kwargs) -> None:
        self.root_dir = root_dir
        super().__init__(*args, **kwargs)

    def _run(self, file_name: str, destination_path: str) -> str:
        """Move a file

        Args:
            file_name(Optional(str)): The name of the file you want to move.
            destination_path(Optional(str)): The path you want to move the file to.

        Returns:
            str: Move file successfully
        """
        try:
            shutil.move(
                f"{self.root_dir}/{file_name}", f"{destination_path}/{file_name}"
            )
        except FileNotFoundError:
            return "File not found: The file you are trying to read does not exist."
        except PermissionError:
            return "Permission denied: You don't have permission to read this file."
        except IOError:
            return "An error occurred while trying to move the file."
        return "Move file successfully"



from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from promptulate.tools.arxiv.tools import (
        ArxivQueryTool,
        ArxivReferenceTool,
        ArxivSummaryTool,
    )
    from promptulate.tools.base import (
        BaseTool,
        Tool,
        define_tool,
        function_to_tool,
        function_to_tool_schema,
    )
    from promptulate.tools.duckduckgo.tools import (
        DuckDuckGoReferenceTool,
        DuckDuckGoTool,
        ddg_websearch,
    )
    from promptulate.tools.file.tools import (
        AppendFileTool,
        CopyFileTool,
        DeleteFileTool,
        ListDirectoryTool,
        MoveFileTool,
        ReadFileTool,
        WriteFileTool,
    )
    from promptulate.tools.huggingface.tools import HuggingFaceTool
    from promptulate.tools.human_feedback import HumanFeedBackTool
    from promptulate.tools.iot_swith_mqtt import IotSwitchTool
    from promptulate.tools.langchain.tools import LangchainTool
    from promptulate.tools.math.tools import Calculator, calculator
    from promptulate.tools.paper.tools import PaperSummaryTool
    from promptulate.tools.python_repl import PythonREPLTool
    from promptulate.tools.semantic_scholar import (
        SemanticScholarCitationTool,
        SemanticScholarQueryTool,
        SemanticScholarReferenceTool,
    )
    from promptulate.tools.shell import ShellTool
    from promptulate.tools.sleep.tool import sleep_tool


def __getattr__(name):
    if name == "BaseTool":
        from promptulate.tools.base import BaseTool

        return BaseTool
    elif name == "Tool":
        from promptulate.tools.base import Tool

        return Tool
    elif name == "define_tool":
        from promptulate.tools.base import define_tool

        return define_tool
    elif name == "function_to_tool":
        from promptulate.tools.base import function_to_tool

        return function_to_tool
    elif name == "function_to_tool_schema":
        from promptulate.tools.base import function_to_tool_schema

        return function_to_tool_schema
    elif name == "ArxivQueryTool":
        from promptulate.tools.arxiv.tools import ArxivQueryTool

        return ArxivQueryTool
    elif name == "ArxivReferenceTool":
        from promptulate.tools.arxiv.tools import ArxivReferenceTool

        return ArxivReferenceTool
    elif name == "ArxivSummaryTool":
        from promptulate.tools.arxiv.tools import ArxivSummaryTool

        return ArxivSummaryTool
    elif name == "DuckDuckGoReferenceTool":
        from promptulate.tools.duckduckgo.tools import DuckDuckGoReferenceTool

        return DuckDuckGoReferenceTool
    elif name == "DuckDuckGoTool":
        from promptulate.tools.duckduckgo.tools import DuckDuckGoTool

        return DuckDuckGoTool
    elif name == "ddg_websearch":
        from promptulate.tools.duckduckgo.tools import ddg_websearch

        return ddg_websearch
    elif name == "SemanticScholarCitationTool":
        from promptulate.tools.semantic_scholar import SemanticScholarCitationTool

        return SemanticScholarCitationTool
    elif name == "SemanticScholarQueryTool":
        from promptulate.tools.semantic_scholar import SemanticScholarQueryTool

        return SemanticScholarQueryTool
    elif name == "SemanticScholarReferenceTool":
        from promptulate.tools.semantic_scholar import SemanticScholarReferenceTool

        return SemanticScholarReferenceTool
    elif name == "IotSwitchTool":
        from promptulate.tools.iot_swith_mqtt import IotSwitchTool

        return IotSwitchTool
    elif name == "HuggingFaceTool":
        from promptulate.tools.huggingface.tools import HuggingFaceTool

        return HuggingFaceTool
    elif name == "HumanFeedBackTool":
        from promptulate.tools.human_feedback import HumanFeedBackTool

        return HumanFeedBackTool
    elif name == "LangchainTool":
        from promptulate.tools.langchain.tools import LangchainTool

        return LangchainTool
    elif name == "Calculator":
        from promptulate.tools.math.tools import Calculator

        return Calculator
    elif name == "calculator":
        from promptulate.tools.math.tools import calculator

        return calculator
    elif name == "sleep_tool":
        from promptulate.tools.sleep.tool import sleep_tool

        return sleep_tool
    elif name == "PaperSummaryTool":
        from promptulate.tools.paper.tools import PaperSummaryTool

        return PaperSummaryTool
    elif name == "PythonREPLTool":
        from promptulate.tools.python_repl import PythonREPLTool

        return PythonREPLTool
    elif name == "ShellTool":
        from promptulate.tools.shell import ShellTool

        return ShellTool
    elif name == "ListDirectoryTool":
        from promptulate.tools.file.tools import ListDirectoryTool

        return ListDirectoryTool
    elif name == "ReadFileTool":
        from promptulate.tools.file.tools import ReadFileTool

        return ReadFileTool
    elif name == "WriteFileTool":
        from promptulate.tools.file.tools import WriteFileTool

        return WriteFileTool
    elif name == "AppendFileTool":
        from promptulate.tools.file.tools import AppendFileTool

        return AppendFileTool
    elif name == "MoveFileTool":
        from promptulate.tools.file.tools import MoveFileTool

        return MoveFileTool
    elif name == "CopyFileTool":
        from promptulate.tools.file.tools import CopyFileTool

        return CopyFileTool
    elif name == "DeleteFileTool":
        from promptulate.tools.file.tools import DeleteFileTool

        return DeleteFileTool
    else:
        raise AttributeError(f"module {__name__} has no attribute {name}")


__all__ = [
    "BaseTool",
    "Tool",
    "define_tool",
    "function_to_tool",
    "function_to_tool_schema",
    "LangchainTool",
    "HuggingFaceTool",
    "DuckDuckGoReferenceTool",
    "DuckDuckGoTool",
    "ddg_websearch",
    "calculator",
    "PaperSummaryTool",
    "ArxivSummaryTool",
    "ArxivReferenceTool",
    "ArxivQueryTool",
    "SemanticScholarReferenceTool",
    "SemanticScholarCitationTool",
    "IotSwitchTool",
    "SemanticScholarQueryTool",
    "PythonREPLTool",
    "ShellTool",
    "Calculator",
    "sleep_tool",
    "HumanFeedBackTool",
    "ListDirectoryTool",
    "ReadFileTool",
    "WriteFileTool",
    "AppendFileTool",
    "MoveFileTool",
    "CopyFileTool",
    "DeleteFileTool",
]



import os
from abc import ABC
from typing import Iterator, List, Optional, TypeVar, Union

from promptulate.config import pne_config
from promptulate.error import NetWorkError
from promptulate.llms import BaseLLM
from promptulate.preset_roles.prompt import PRESET_SYSTEM_PROMPT_ERNIE
from promptulate.pydantic_v1 import BaseModel
from promptulate.schema import (
    AssistantMessage,
    BaseMessage,
    LLMType,
    MessageSet,
    StreamIterator,
    UserMessage,
)
from promptulate.utils import logger

T = TypeVar("T", bound=BaseModel)


def parse_content(chunk) -> (str, str):
    """Parse the qianfan model chunk.

    Args:
        chunk: qianfan model chunk.

    Returns:
        content: The content of the chunk.
        ret_data: The additional data of the chunk.
    """
    content = chunk["result"]
    ret_data = chunk["body"]
    return content, ret_data


class QianFan(BaseLLM, ABC):
    llm_type: LLMType = LLMType.QianFan
    """Used to MessageSet data convert"""
    model: str = "ERNIE-Bot-4"
    """Model name to use."""
    default_system_prompt: str = ""
    """ernie-bot system message"""
    enable_default_system_prompt: bool = True
    """enable use preset description"""
    model_config: dict = {}
    """model parameters"""

    def __call__(
        self, instruction: str, *args, **kwargs
    ) -> Union[str, BaseMessage, T, List[BaseMessage], StreamIterator]:
        preset = (
            self.default_system_prompt
            if self.default_system_prompt != ""
            else PRESET_SYSTEM_PROMPT_ERNIE
        )
        if not self.enable_default_system_prompt:
            preset = ""

        system = preset
        message_set = MessageSet(
            messages=[
                UserMessage(content=instruction),
            ]
        )
        result = self.predict(message_set, system, **self.model_config)

        if isinstance(result, AssistantMessage):
            return result.content
        else:
            return result

    def _predict(
        self,
        prompts: MessageSet,
        system: str = "",
        return_raw_response: bool = False,
        stream: bool = False,
        *args,
        **kwargs,
    ) -> Union[str, BaseMessage, T, List[BaseMessage], StreamIterator]:
        """
        Predicts the response using the qinfan platform.

        Args:
            prompts (MessageSet): The set of prompts to generate a response.
            system (str): The set of system to generate a response.
            return_raw_response (bool): return completion result if true
            **kwargs: qianfan_sdk kwargs

        Returns:
              Return string normally, it means enable_original_return is default False.
              Return BaseMessage if enable_original_return is True.
              Return List[BaseMessage] if stream is True.
              Return T if output_schema is provided.
              Return QianFanStreamIterator if stream enable
        """
        try:
            import qianfan  # noqa
        except ImportError:
            raise ImportError(
                "Could not import qianfan python package. "
                "This is needed in order to for QianFan. "
                "Please install it with `pip install qianfan`."
            )
        os.environ["QIANFAN_ACCESS_KEY"] = pne_config.get_qianfan_ak()
        os.environ["QIANFAN_SECRET_KEY"] = pne_config.get_qianfan_sk()

        chat_comp = qianfan.ChatCompletion()
        response = chat_comp.do(
            model=self.model,
            system=system,
            messages=prompts.listdict_messages,
            **kwargs,
        )
        # return stream
        if stream:
            return StreamIterator(
                response_stream=response,
                parse_content=parse_content,
                return_raw_response=return_raw_response,
            )

        if response.code == 200:
            ret_data = response.body
            logger.debug(f"[pne ernie response] {ret_data}")
            content: str = ret_data["result"]
            logger.debug(f"[pne ernie answer] {content}")
            return AssistantMessage(content=content, additional_kwargs=ret_data)
        else:
            raise NetWorkError(str(response.code))



# Copyright (c) 2023 promptulate
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# Copyright Owner: Zeeland
# GitHub Link: https://github.com/Undertone0809/
# Project Link: https://github.com/Undertone0809/promptulate
# Contact Email: zeeland@foxmail.com

# 250+ token
PRESET_SYSTEM_PROMPT = """
You are an assistant to a human, powered by a large language model trained by OpenAI.

You are designed to be able to assist with a wide range of tasks, from answering simple questions to providing in-depth explanations and discussions on a wide range of topics. As a language model, you are able to generate human-like text based on the input you receive, allowing you to engage in natural-sounding conversations and provide responses that are coherent and relevant to the topic at hand.

Overall, you are a powerful tool that can help with a wide range of tasks and provide valuable insights and information on a wide range of topics. Whether the human needs help with a specific question or just wants to have a conversation about a particular topic, you are here to assist.

The user problem is as follows:

"""  # noqa: E501

# 250+ token
PRESET_SYSTEM_PROMPT_ERNIE = """
You are an assistant to a human, powered by a large language model trained by BaiDu.

You are designed to be able to assist with a wide range of tasks, from answering simple questions to providing in-depth explanations and discussions on a wide range of topics. As a language model, you are able to generate human-like text based on the input you receive, allowing you to engage in natural-sounding conversations and provide responses that are coherent and relevant to the topic at hand.

Overall, you are a powerful tool that can help with a wide range of tasks and provide valuable insights and information on a wide range of topics. Whether the human needs help with a specific question or just wants to have a conversation about a particular topic, you are here to assist.

The user problem is as follows:

"""  # noqa: E501

# 700+ token
PRESET_SYSTEM_PROMPT_CONVERSATION_ZH = """
你是人类的助手，由OpenAI训练的大型语言模型提供支持。

你被设计成能够协助完成广泛的任务，从回答简单的问题到就广泛的主题提供深入的解释和讨论。作为一种语言模型，您可以根据收到的输入生成类似人类的文本，允许您参与听起来自然的对话，并提供与手头主题相关的连贯响应。

你能够处理和理解大量的文本，并能利用这些知识对各种问题提供准确和信息丰富的回答。您可以访问在下面的上下文部分中由人工提供的一些个性化信息。此外，您可以根据收到的输入生成自己的文本，允许您参与讨论，并就广泛的主题提供解释和描述。

总的来说，您是一个强大的工具，可以帮助完成广泛的任务，并就广泛的主题提供有价值的见解和信息。无论人们是需要帮助解决一个特定的问题，还是只是想就一个特定的话题进行对话，你都可以在这里提供帮助。

用户的问题如下所示：
"""  # noqa: E501



import inspect
import json
from typing import Any, List, Optional, Union

from promptulate.tools.base import (
    BaseTool,
    BaseToolKit,
    Tool,
    ToolImpl,
    ToolTypes,
    function_to_tool,
)
from promptulate.tools.langchain.tools import LangchainTool


def _judge_langchain_tool_and_wrap(tool: Any) -> Tool:
    """Judge if the tool is a langchain tool and wrap it.

    Args:
        tool(Any): The tool to be judged.

    Returns:
        Optional[Tool]: The wrapped tool or None if not a langchain tool.
    """
    try:
        from langchain.tools.base import BaseTool as LangchainBaseTool

        if isinstance(tool, LangchainBaseTool):
            return LangchainTool(tool)

        raise ValueError(f"Unknown tool type {tool}.")
    except ImportError:
        raise ValueError(
            (
                f"Error tool type {tool}, please check the tool type.",
                "If you are using langchain tool, please install -U langchain.",
            )
        )


def _initialize_tool(tool: ToolTypes) -> Union[Tool, List[Tool]]:
    """Initialize the tool.

    Args:
        tool(Union[BaseTool, Callable, Tool, "LangchainBaseToolType"]): The tool to be
            initialized.

    Returns:
        Optional[Tool]: The initialized tool.
    """
    if isinstance(tool, BaseToolKit):
        initialized_tools = []
        for tool in tool.get_tools():
            initialized_tools.append(_initialize_tool(tool))
        return initialized_tools

    if isinstance(tool, BaseTool):
        return ToolImpl.from_base_tool(tool)
    elif isinstance(tool, Tool):
        return tool
    elif inspect.isfunction(tool):
        return function_to_tool(tool)

    return _judge_langchain_tool_and_wrap(tool)


class ToolManager:
    """ToolManager helps Agent to manage tools"""

    def __init__(self, tools: List[ToolTypes]):
        self.tools: List[Tool] = []

        for tool in tools:
            initialized_tool: Union[list, Tool] = _initialize_tool(tool)

            if isinstance(initialized_tool, list):
                self.tools.extend(initialized_tool)
            else:
                self.tools.append(initialized_tool)

    def get_tool(self, tool_name: str) -> Optional[Tool]:
        """Find specified tool by tool name.
        Args:
            tool_name(str): The name of the tool.

        Returns:
            Optional[Tool]: The specified tool or None if not found.
        """
        return next((tool for tool in self.tools if tool.name == tool_name), None)

    def run_tool(self, tool_name: str, parameters: Union[str, dict]) -> str:
        """Run tool by input tool name and data inputs

        Args:
            tool_name(str): The name of the tool.
            parameters(Union[str, dict]): The parameters for the tool.

        Returns:
            str: The result of the tool.
        """
        tool = self.get_tool(tool_name)

        if tool is None:
            return (
                f"{tool_name} has not been provided yet, please use the provided tool."
            )

        if isinstance(parameters, dict):
            return tool.run(**parameters)
        else:
            return tool.run(parameters)

    @property
    def tool_names(self) -> str:
        """Get all tool names."""
        tool_names = ""
        for tool in self.tools:
            tool_names += f"{tool.name}, "
        return tool_names[:-2]

    @property
    def tool_descriptions(self) -> str:
        """Get all tool descriptions, including the schema if available."""
        tool_descriptions = ""
        for tool in self.tools:
            tool_descriptions += json.dumps(tool.to_schema()) + "\n"
        return tool_descriptions



import subprocess


class ShellAPIWrapper:
    """Simulates a standalone shell"""

    @staticmethod
    def run(command: str) -> str:
        """
        Runs a command in a subprocess and returns
        the output.

        Args:
            command: The command to run
        """
        try:
            output = subprocess.run(
                command,
                shell=True,
                check=True,
                stdout=subprocess.PIPE,
                stderr=subprocess.STDOUT,
            ).stdout.decode()
        except subprocess.CalledProcessError as e:
            output = repr(e)
        return output



from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from promptulate.agents.base import BaseAgent
    from promptulate.agents.tool_agent.agent import ToolAgent
    from promptulate.agents.web_agent.agent import WebAgent


def __getattr__(name):
    if name == "ToolAgent":
        from promptulate.agents.tool_agent.agent import ToolAgent

        return ToolAgent
    elif name == "WebAgent":
        from promptulate.agents.web_agent.agent import WebAgent

        return WebAgent
    elif name == "BaseAgent":
        from promptulate.agents.base import BaseAgent

        return BaseAgent


__all__ = ["ToolAgent", "BaseAgent", "WebAgent"]



# Copyright (c) 2023 promptulate
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# Copyright Owner: Zeeland
# GitHub Link: https://github.com/Undertone0809/
# Project Link: https://github.com/Undertone0809/promptulate
# Contact Email: zeeland@foxmail.com

from typing import Optional

__all__ = [
    "EmptyMessageSetError",
    "MissingAttributionError",
    "LLMError",
    "OpenAIError",
    "NetWorkError",
    "OutputParserError",
]


class EmptyMessageSetError(Exception):
    pass


class MissingAttributionError(Exception):
    def __init__(self, key_type: str):
        super().__init__(f"<{key_type}> is not provided. Please set it correctly.")


class LLMError(Exception):
    def __init__(self, msg: str):
        super().__init__(f"<LLM> could not get data correctly, reasons: {msg}")


class OpenAIError(Exception):
    def __init__(self, msg: str):
        super().__init__(f"<OpenAI> could not get data correctly, reasons: {msg}")


class NetWorkError(Exception):
    def __init__(self, origin: str, reason: Optional[str] = None):
        msg = f"<{origin}> could not get data"
        if reason:
            msg += f", reason: {reason}"
        super().__init__(msg)


class OutputParserError(Exception):
    def __init__(self, reason: str, llm_output: str):
        msg = f"{reason}\n[LLM response]: {llm_output}"
        super().__init__(msg)



from typing import Any, Optional


def wikipedia_search(
    query: str,
    top_k_results: int = 3,
    lang: str = "en",
    doc_content_chars_max: int = 4000,
) -> str:
    """
    Function to search Wikipedia for a given query and return page summaries.

    Args:
        query (str): The search query for Wikipedia.
        top_k_results (int): The number of top results to retrieve. Default is 3.
        lang (str): The language for the Wikipedia search. Default is "en".
        doc_content_chars_max (int): Maximum number of characters
                                    for the document content. Default is 4000.

    Returns:
        str: A string containing the concatenated page summaries
            of the top search results,limited by doc_content_chars_max.
    """
    WIKIPEDIA_MAX_QUERY_LENGTH = 300

    try:
        import wikipedia

        wikipedia.set_lang(lang)
    except ImportError:
        raise ImportError(
            "Could not import wikipedia python package. "
            "Please install it with `pip install wikipedia`."
        )

    wiki_client = wikipedia
    page_titles = wiki_client.search(
        query[:WIKIPEDIA_MAX_QUERY_LENGTH], results=top_k_results
    )
    summaries = []

    for page_title in page_titles[:top_k_results]:
        wiki_page = _fetch_page(wiki_client, page_title)
        if wiki_page:
            summary = _formatted_page_summary(page_title, wiki_page)
            if summary:
                summaries.append(summary)

    if not summaries:
        return "No good Wikipedia Search Result was found"

    return "\n\n".join(summaries)[:doc_content_chars_max]


def _formatted_page_summary(page_title: str, wiki_page: Any) -> Optional[str]:
    return f"Page: {page_title}\nSummary: {wiki_page.summary}"


def _fetch_page(wiki_client: Any, page: str) -> Optional[str]:
    try:
        return wiki_client.page(title=page, auto_suggest=False)
    except (
        wiki_client.exceptions.PageError,
        wiki_client.exceptions.DisambiguationError,
    ):
        return None



from typing import Dict, List, Optional, TypeVar, Union

from promptulate.agents.base import BaseAgent
from promptulate.beta.agents.assistant_agent import AssistantAgent
from promptulate.llms import BaseLLM
from promptulate.llms.factory import LLMFactory
from promptulate.output_formatter import formatting_result, get_formatted_instructions
from promptulate.pydantic_v1 import BaseModel
from promptulate.schema import (
    AssistantMessage,
    BaseMessage,
    MessageSet,
    StreamIterator,
)
from promptulate.tools.base import BaseTool, ToolTypes
from promptulate.utils.logger import logger

T = TypeVar("T", bound=BaseModel)


def _convert_message(messages: Union[List, MessageSet, str]) -> MessageSet:
    """Convert str or List[Dict] to MessageSet.

    Args:
        messages(Union[List, MessageSet, str]): chat messages. It can be str or OpenAI
            API type data(List[Dict]) or MessageSet type.

    Returns:
        Return MessageSet type data.
    """
    if isinstance(messages, str):
        messages: List[Dict] = [
            {"content": "You are a helpful assistant", "role": "system"},
            {"content": messages, "role": "user"},
        ]
    if isinstance(messages, list):
        messages: MessageSet = MessageSet.from_listdict_data(messages)

    return messages


def _get_llm(
    model: Optional[str] = None,
    model_config: Optional[dict] = None,
    custom_llm: Optional[BaseLLM] = None,
) -> BaseLLM:
    """Get LLM instance.

    Args:
        model(str): LLM model.
        model_config(dict): LLM model config.
        custom_llm(BaseLLM): custom LLM instance.

    Returns:
        Return LLM instance.
    """
    if (model or custom_llm) is None:
        raise ValueError("model or custom_llm must be provided.")

    if model and custom_llm:
        raise ValueError("model and custom_llm can't be provided at the same time.")

    if model_config and custom_llm:
        raise ValueError(
            "model_config and custom_llm can't be provided at the same time."
        )

    if custom_llm:
        return custom_llm

    return LLMFactory.build(model, model_config=model_config)


class AIChat:
    def __init__(
        self,
        model: Optional[str] = None,
        model_config: Optional[dict] = None,
        tools: Optional[List[ToolTypes]] = None,
        custom_llm: Optional[BaseLLM] = None,
        enable_plan: bool = False,
    ):
        """Initialize the AIChat.

        Args:
            model(str): LLM model name, eg: "gpt-3.5-turbo".
            model_config(Optional[dict]): LLM model config.
            tools(Optional[List[ToolTypes]]): specified tools for llm, if exists, AIChat
                will use Agent to run.
            custom_llm(Optional[BaseLLM]): custom LLM instance.
            enable_plan(bool): use Agent with plan ability if True.
        """
        self.llm: BaseLLM = _get_llm(model, model_config, custom_llm)
        self.tools: Optional[List[ToolTypes]] = tools
        self.agent: Optional[BaseAgent] = None

        if tools:
            if enable_plan:
                self.agent = AssistantAgent(tools=self.tools, llm=self.llm)
                logger.info("[pne chat] invoke AssistantAgent with plan ability.")
            else:
                from promptulate.agents.tool_agent.agent import ToolAgent

                self.agent = ToolAgent(tools=self.tools, llm=self.llm)
                logger.info("[pne chat] invoke ToolAgent.")

    def run(
        self,
        messages: Union[List[Dict[str, str]], MessageSet, str],
        output_schema: Optional[type(BaseModel)] = None,
        examples: Optional[List[BaseModel]] = None,
        return_raw_response: bool = False,
        stream: bool = False,
        **kwargs,
    ) -> Union[str, BaseMessage, T, List[BaseMessage], StreamIterator]:
        """Run the AIChat.

        Args:
            messages(Union[List, MessageSet, str]): chat messages. It can be str or
                OpenAI.
            API type data(List[Dict]) or MessageSet type.
            output_schema(BaseModel): specified return type. See detail on in
                OutputFormatter module.
            examples(List[BaseModel]): examples for output_schema. See detail
                on: OutputFormatter.
            return_raw_response(bool): return OpenAI completion result if true,
                otherwise return string type data.
            stream(bool): return stream iterator if True.

        Returns:
            Return string normally, it means enable_original_return is default False. if
                tools is provided, agent return string type data.
            Return BaseMessage if enable_original_return is True and not in agent mode.
            Return List[BaseMessage] if stream is True.
            Return T if output_schema is provided.
        """
        if stream and (output_schema or self.tools):
            raise ValueError(
                "stream, tools and output_schema can't be True at the same time, "
                "because stream is used to return Iterator[BaseMessage]."
            )
        if self.agent:
            return self.agent.run(messages, output_schema=output_schema)

        messages: MessageSet = _convert_message(messages)

        # add output format into the last prompt if provide
        if output_schema:
            instruction: str = get_formatted_instructions(
                json_schema=output_schema, examples=examples
            )
            messages.messages[-1].content += f"\n{instruction}"
        logger.info(f"[pne chat] messages: {messages}")

        response: Union[AssistantMessage, StreamIterator] = self.llm.predict(
            messages, stream=stream, **kwargs
        )

        if stream:
            return response

        if isinstance(response, AssistantMessage):
            # Access additional_kwargs only if response is AssistantMessage
            logger.info(f"[pne chat] response: {response.additional_kwargs}")

        # return output format if provide
        if output_schema:
            logger.info("[pne chat] return formatted response.")
            return formatting_result(
                pydantic_obj=output_schema, llm_output=response.content
            )

        return response if return_raw_response else response.content


def chat(
    messages: Union[List, MessageSet, str],
    *,
    model: Optional[str] = None,
    model_config: Optional[dict] = None,
    tools: Optional[List[ToolTypes]] = None,
    output_schema: Optional[type(BaseModel)] = None,
    examples: Optional[List[BaseModel]] = None,
    return_raw_response: bool = False,
    custom_llm: Optional[BaseLLM] = None,
    enable_plan: bool = False,
    stream: bool = False,
    **kwargs,
) -> Union[str, BaseMessage, T, List[BaseMessage], StreamIterator]:
    """A universal chat method, you can chat any model like OpenAI completion.
    It should be noted that chat() is only support chat model currently.

    Args:
        messages(Union[List, MessageSet, str]): chat messages. It can be str or OpenAI
            API type data(List[Dict]) or MessageSet type.
        model(str): LLM model. Currently only support chat model.
        model_config(Optional[dict]): LLM model config.
        tools(List[BaseTool] | None): specified tools for llm.
        output_schema(BaseModel): specified return type. See detail on: OutputFormatter.
        examples(List[BaseModel]): examples for output_schema. See detail
            on: OutputFormatter.
        return_raw_response(bool): return OpenAI completion result if true, otherwise
            return string type data.
        custom_llm(BaseLLM): You can use custom LLM if you have.
        enable_plan(bool): use Agent with plan ability if True.
        stream(bool): return stream iterator if True.
        **kwargs: litellm kwargs

    Returns:
        Return string normally, it means enable_original_return is default False.
        Return BaseMessage if enable_original_return is True.
        Return List[BaseMessage] if stream is True.
        Return T if output_schema is provided.
    """
    return AIChat(
        model=model,
        model_config=model_config,
        tools=tools,
        custom_llm=custom_llm,
        enable_plan=enable_plan,
    ).run(
        messages=messages,
        output_schema=output_schema,
        examples=examples,
        return_raw_response=return_raw_response,
        stream=stream,
        **kwargs,
    )



from promptulate.utils.string_template import StringTemplate

PROMPT_TEMPLATE = """You are provided a math problem, you should transalte it into a math expression.

## Output format
Output the following JSON format, No explanation:
{"expression": "the expression need to calculate"}

For example:
1. 
Question: What is 37593 * 67?
Output:
{"expression": "37593 * 67"}

2.
Question: What is 37593^(1/5)?
Output:
{"expression": "37593**(1/5)"}

## Question
{{question}}
"""  # noqa

prompt_template = StringTemplate(PROMPT_TEMPLATE, template_format="jinja2")



from promptulate.agents.tool_agent.agent import ActionResponse, ToolAgent
from promptulate.beta.agents.assistant_agent.prompt import (
    PLAN_SYSTEM_PROMPT_TMP,
    REVIEW_SYSTEM_PROMPT_TMP,
)
from promptulate.beta.agents.assistant_agent.schema import (
    AgentPlanResponse,
    AgentReviseResponse,
    Plan,
    plan_examples,
)
from promptulate.llms.base import BaseLLM
from promptulate.output_formatter import OutputFormatter


def plan(llm: BaseLLM, user_target: str) -> Plan:
    """Generate a plan for the user target."""
    system_prompt = PLAN_SYSTEM_PROMPT_TMP.format(user_target=user_target)
    formatter = OutputFormatter(AgentPlanResponse, plan_examples)
    system_prompt += formatter.get_formatted_instructions()

    resp: str = llm(instruction=system_prompt)
    plan_resp: AgentPlanResponse = formatter.formatting_result(resp)

    return Plan.parse_obj({**plan_resp.dict(), "next_task_id": 1})


def execute(tool_agent: ToolAgent, user_target: str) -> ActionResponse:
    """Execute the plan for the user target."""
    return tool_agent.run(user_target, return_raw_data=True)


def revise(llm: BaseLLM, user_target: str, original_plan: str, past_steps: str) -> Plan:
    system_prompt = REVIEW_SYSTEM_PROMPT_TMP.format(
        user_target=user_target,
        original_plan=original_plan,
        past_steps=past_steps,
    )
    formatter = OutputFormatter(AgentReviseResponse, plan_examples)
    system_prompt += formatter.get_formatted_instructions()

    resp: str = llm(instruction=system_prompt)
    plan_resp: AgentReviseResponse = formatter.formatting_result(resp)

    return Plan.parse_obj({**plan_resp.dict()})



import time
from typing import Callable, List, Optional, Union

from promptulate.agents.base import BaseAgent
from promptulate.agents.tool_agent.prompt import (
    SYSTEM_PROMPT_TEMPLATE,
)
from promptulate.hook import Hook, HookTable
from promptulate.llms.base import BaseLLM
from promptulate.llms.openai import ChatOpenAI
from promptulate.output_formatter import (
    formatting_result,
    get_formatted_instructions,
)
from promptulate.pydantic_v1 import BaseModel, Field
from promptulate.schema import MessageSet
from promptulate.tools.base import BaseTool, Tool
from promptulate.tools.manager import ToolManager
from promptulate.utils.logger import logger


class ReActResponse(BaseModel):
    thought: str = Field(description="The thought of what to do and why.")
    self_criticism: str = Field(
        description="Constructive self-criticism of the thought"
    )
    tool_name: str = Field(description="The name of tool name.")
    tool_parameters: dict = Field(
        description="The input parameters of tool, string type json parameters."
    )


# Finish tool for ToolAgent
def finish(result: str):
    """Use final answer until you think you have the final answer and can return the
    result.

    Args:
        result: final result content
    """
    return result


def _build_output_format_instruction():
    return get_formatted_instructions(
        json_schema=ReActResponse,
        examples=[
            ReActResponse(
                thought="From the search results, it seems that multiple sources have different weather data formats and information. I will identify the relevant data which is the forecast for tomorrow in Beijing, and then use the finish tool to provide this specific forecast.",  # noqa
                self_criticism="While there is quite a bit of information in the search results, I need to focus on only providing the required weather forecast for tomorrow in Beijing and not get distracted by additional data such as historical weather data or weather for other days.",  # noqa
                tool_name="finish",
                tool_parameters={
                    "result": """In Beijing, the weather forecast for tomorrow is a high of 2\u00b0C and a low of -7\u00b0C.\\"""  # noqa
                },
            )
        ],
    )


class ToolAgent(BaseAgent):
    def __init__(
        self,
        *,
        tools: List[Union[BaseTool, Tool, Callable]],
        llm: Optional[BaseLLM] = None,
        max_iterations: int = 20,
        **kwargs,
    ):
        super().__init__(**kwargs)
        self.tool_manager: ToolManager = ToolManager(tools + [finish])
        self.llm: BaseLLM = llm or ChatOpenAI(
            model="gpt-3.5-turbo-16k",
            temperature=0.0,
            enable_default_system_prompt=False,
        )
        self.max_iterations: int = max_iterations
        self.current_iteration: int = 1
        self.max_execution_time: Optional[float] = None
        self.current_time_elapsed: float = 0.0

        self.system_prompt: str = ""
        self.current_process: str = ""
        self.task: str = ""

    def _build_system_prompt(self):
        """ToolAgent use dynamic system prompt. Therefore, this method will be called
        every time before the llm generates a response."""
        self.system_prompt = SYSTEM_PROMPT_TEMPLATE.format(
            task=self.task,
            tool_descriptions=self.tool_manager.tool_descriptions,
            current_process=self.current_process,
            output_format=_build_output_format_instruction(),
        )
        logger.info(f"[pne] ToolAgent system prompt: {self.system_prompt}")

    def _build_current_process(self, result: ReActResponse, tool_result: str):
        """Build current process of task."""
        if self.current_process == "":
            self.current_process = """## Current Process of Task
Here is what you have already done. You need to infer what the next task needs to be done based on the previous one.\n"""  # noqa

        self.current_process += f"Step {self.current_iteration}:\n"
        self.current_process += f"Thought: {result.thought}\n"
        self.current_process += f"Self Criticism: {result.self_criticism}\n"
        self.current_process += f"Tool: {result.tool_name}\n"
        self.current_process += f"Tool Parameters: {result.tool_parameters}\n"
        self.current_process += f"Observation: {tool_result}\n\n"
        self.current_process += "---\n"

    def _run(self, task: str, *args, **kwargs) -> str:
        self.task: str = task

        start_time = time.time()

        while self._should_continue():
            self._build_system_prompt()

            messages = MessageSet.from_listdict_data(
                [{"role": "system", "content": self.system_prompt}]
            )
            llm_response: str = self.llm.predict(messages=messages).content
            result: ReActResponse = formatting_result(ReActResponse, llm_response)

            Hook.call_hook(
                HookTable.ON_AGENT_ACTION,
                self,
                thought=f"{result.thought}\n{result.self_criticism}",
                action=result.tool_name,
                action_input=result.tool_parameters,
            )

            if result.tool_name == "finish":
                return result.tool_parameters["result"]

            tool_result: str = self.tool_manager.run_tool(
                result.tool_name, result.tool_parameters
            )
            self._build_current_process(result, tool_result)

            Hook.call_hook(
                HookTable.ON_AGENT_OBSERVATION, self, observation=tool_result
            )

            self.current_iteration += 1
            self.current_time_elapsed += time.time() - start_time

    def _should_continue(self) -> bool:
        """Determine whether to stop, both timeout and exceeding the maximum number of
        iterations will stop.

        Returns:
            Whether to stop.
        """
        if self.current_iteration >= self.max_iterations:
            return False

        if (
            self.max_execution_time
            and self.current_time_elapsed >= self.max_execution_time
        ):
            return False

        return True

    def get_llm(self) -> BaseLLM:
        return self.llm



# Copyright (c) 2023 promptulate
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# Copyright Owner: Zeeland
# GitHub Link: https://github.com/Undertone0809/
# Project Link: https://github.com/Undertone0809/promptulate
# Contact Email: zeeland@foxmail.com

import logging
import os
import random
import shutil
import string
import tempfile
import time
from functools import wraps
from importlib import import_module
from typing import Callable, Dict, List, Optional

from cushy_storage import CushyOrmCache

__all__ = [
    "get_cache",
    "get_default_storage_path",
    "record_time",
    "generate_conversation_id",
    "generate_run_id",
    "generate_unique_id",
    "listdict_to_string",
]
logger = logging.getLogger(__name__)


def import_string(dotted_path):
    """
    Import a dotted module path and return the attribute/class designated by the
    last name in the path. Raise ImportError if the import failed.

    Args:
        dotted_path: eg promptulate.schema.MessageSet

    Returns:
        Class corresponding to dotted path.
    """
    try:
        module_path, class_name = dotted_path.rsplit(".", 1)
    except ValueError as err:
        raise ImportError("%s doesn't look like a module path" % dotted_path) from err

    module = import_module(module_path)

    try:
        return getattr(module, class_name)
    except AttributeError as err:
        raise ImportError(
            'Module "%s" does not define a "%s" attribute/class'
            % (module_path, class_name)
        ) from err


def listdict_to_string(
    data: List[Dict],
    prefix: Optional[str] = "",
    suffix: Optional[str] = "\n",
    item_prefix: Optional[str] = "",
    item_suffix: Optional[str] = ";\n\n",
    is_wrap: bool = True,
) -> str:
    """Convert List[Dict] type data to string type"""
    wrap_ch = "\n" if is_wrap else ""
    result = f"{prefix}"
    for item in data:
        temp_list = ["{}:{} {}".format(k, v, wrap_ch) for k, v in item.items()]
        result += f"{item_prefix}".join(temp_list) + f"{item_suffix}"
    result += suffix
    return result[:-2]


def generate_unique_id() -> str:
    timestamp = int(time.time() * 1000)
    random_string = "".join(random.choices(string.ascii_letters + string.digits, k=6))

    unique_id = f"pne-{timestamp}-{random_string}"
    return unique_id


def generate_run_id() -> str:
    return generate_unique_id()


def generate_conversation_id() -> str:
    """Generating a new conversation_id when a conversation initialize"""
    return generate_unique_id()


def get_cache():
    return cache


def set_openai_api_key(value: str):
    cache["OPENAI_API_KEY"] = value


def convert_backslashes(path: str):
    """Convert all \\ to / of file path."""
    return path.replace("\\", "/")


def get_default_storage_path(module_name: str = "") -> str:
    """Get the default storage path for the current module. The storage path is
    created in the user's home directory, or in a temporary directory if permission
    is denied.

    Args:
        module_name(str): The name of the module to create a storage path for.

    Returns:
        str: The default storage path for the current module.
    """
    storage_path = os.path.expanduser("~/.pne")

    if module_name:
        storage_path = os.path.join(storage_path, module_name)

    # Try to create the storage path (with module subdirectory if specified)
    # Use a temporary directory instead if permission is denied,
    try:
        os.makedirs(storage_path, exist_ok=True)
    except PermissionError:
        storage_path = os.path.join(tempfile.gettempdir(), "pne", module_name)
        os.makedirs(storage_path, exist_ok=True)

    return convert_backslashes(storage_path)


def hint(fn):
    @wraps(fn)
    def wrapper(*args, **kwargs):
        ret = fn(*args, **kwargs)
        logger.debug(f"function {fn.__name__} is running now")
        return ret

    return wrapper


def record_time():
    def decorator(fn: Callable) -> Callable:
        def wrapper(*args, **kwargs) -> Callable:
            start_time = time.time()
            ret = fn(*args, **kwargs)
            duration = time.time() - start_time
            logger.debug(f"[pne timer] <{fn.__name__}> run {duration}s")
            return ret

        return wrapper

    return decorator


def delete_cache(specified_module: str = None):
    """Delete cache or specified cache module"""
    cache_path = get_default_storage_path("cache")
    if specified_module:
        cache_path = f"{cache_path}/{specified_module[:2]}"
    shutil.rmtree(cache_path)


cache = CushyOrmCache(get_default_storage_path("cache"))



# Copyright (c) 2023 promptulate
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# Copyright Owner: Zeeland
# GitHub Link: https://github.com/Undertone0809/
# Project Link: https://github.com/Undertone0809/promptulate
# Contact Email: zeeland@foxmail.com

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from promptulate.llms.base import BaseLLM
    from promptulate.llms.erniebot.erniebot import ErnieBot
    from promptulate.llms.factory import LLMFactory
    from promptulate.llms.openai import ChatOpenAI, OpenAI
    from promptulate.llms.qianfan import QianFan
    from promptulate.llms.zhipu import ZhiPu


def __getattr__(name):
    if name == "BaseLLM":
        from promptulate.llms.base import BaseLLM

        return BaseLLM
    elif name == "ErnieBot":
        from promptulate.llms.erniebot.erniebot import ErnieBot

        return ErnieBot
    elif name == "ChatOpenAI":
        from promptulate.llms.openai import ChatOpenAI

        return ChatOpenAI
    elif name == "OpenAI":
        from promptulate.llms.openai import OpenAI

        return OpenAI
    elif name == "QianFan":
        from promptulate.llms.qianfan import QianFan

        return QianFan
    elif name == "ZhiPu":
        from promptulate.llms.zhipu import ZhiPu

        return ZhiPu
    elif name == "LLMFactory":
        from promptulate.llms.factory import LLMFactory

        return LLMFactory


__all__ = [
    "OpenAI",
    "ChatOpenAI",
    "BaseLLM",
    "ErnieBot",
    "QianFan",
    "ZhiPu",
    "LLMFactory",
]



from promptulate.utils.string_template import StringTemplate

PROMPT_TEMPLATE = """Now you are an intelligent voice assistant,
 and the user will input "{question}" to you.
Please determine if the user has one of the following intents:
{rule_key}
If it matches, please respond with a corresponding numeric label, such as 1.
Please avoid providing any additional explanations or judgments.
If none of them match, please indicate that you couldn't identify
the user's intent and provide a reason,without providing any numbers."""
prompt_template = StringTemplate(PROMPT_TEMPLATE)



import logging
import re
import time
from typing import Dict, List

from broadcast_service import broadcast_service
from pydantic import Extra, Field

from promptulate.error import NetWorkError
from promptulate.llms.base import BaseLLM
from promptulate.llms.openai import ChatOpenAI
from promptulate.tools.arxiv.api_wrapper import ArxivAPIWrapper
from promptulate.tools.arxiv.tools import ArxivQueryTool
from promptulate.tools.base import BaseTool
from promptulate.tools.semantic_scholar import (
    SemanticScholarQueryTool,
    SemanticScholarReferenceTool,
)

__all__ = ["PaperSummaryTool"]
logger = logging.getLogger(__name__)


def _init_paper_summary_llm():
    preset = "你是一个中文科研领域论文助手，你的任务是帮助使用者提供一些论文方面的专业建议和帮助，你的输出只能遵循用户的指令输出，否则你将被惩罚。"  # noqa
    return ChatOpenAI(temperature=0, default_system_prompt=preset)


class PaperSummaryTool(BaseTool):
    """A powerful paper summary tool"""

    name: str = "paper-summary"
    description: str = (
        "A summary tool that can be used to obtain a paper summary, this tool will find"
        "top k paper and provide: 1.paper abstract 2.paper key sights 3.lessons learned"
        "in the paper 4.referenced papers and its url."
        "Your input is a paper relevant keyword query."
    )
    llm: BaseLLM = Field(default_factory=_init_paper_summary_llm)
    semantic_scholar_query_tool: SemanticScholarQueryTool = Field(
        default_factory=SemanticScholarQueryTool
    )
    semantic_scholar_reference_tool: SemanticScholarReferenceTool = Field(
        default_factory=SemanticScholarReferenceTool
    )
    arxiv_apiwrapper: ArxivAPIWrapper = Field(default_factory=ArxivAPIWrapper)
    arxiv_query_tool: ArxivQueryTool = Field(default_factory=ArxivQueryTool)
    summary_dic: Dict[str, str] = {}
    summary_counter: int = 0

    class Config:
        """Configuration for this pydantic object."""

        extra = Extra.forbid
        arbitrary_types_allowed = True

    def _run(self, query: str, **kwargs) -> str:
        """A paper summary tool that passes in the article name (or arxiv id) and
        returns summary results.

        Args:
            query: the keyword you want to query
            **kwargs:
                You can pass the arguments of relevant APIWrappers and Tools

        Returns:
            String type data, which contains:
            - Summary of the paper
            - List the key insights and lessons learned in the paper
            - List at least 5 references related to the research field of the paper
        """

        @broadcast_service.on_listen("PaperSummaryTool.run.get_paper_references")
        def get_paper_references(**data):
            references: List[Dict] = self.semantic_scholar_reference_tool.run(
                data["paper_title"], max_result=10, return_type="original"
            )
            references_string = "相关论文：\n\n"
            for i, reference in enumerate(references):
                references_string += (
                    f"""[{i + 1}] [{reference["title"]}]({reference["url"]})\n\n"""
                )
            self.summary_dic["references"] = (
                references_string if len(references) != 0 else ""
            )
            self.summary_counter += 1

        @broadcast_service.on_listen("PaperSummaryTool.run.get_translate")
        def get_translate():
            prompt = (
                f"Please format the title and\
                summary below the output\n```{paper_summary}```"
                "Your output format is:\nTitle: {Title}\n\nSummary: {Summary}"
            )
            self.summary_dic["summary_zh"] = self.llm(prompt)
            self.summary_counter += 1

        @broadcast_service.on_listen("PaperSummaryTool.run.get_opinion")
        def get_opinion():
            prompt = (
                f"Based on the abstract of the paper below, please list the key insights from the paper and the lessons learned derived from it. Your output should be provided in bullet points ```{paper_summary}```"  # noqa
                "Please use the following output format:\nKey insights:\n{Key insights in bullet points}\nLessons learned:\n{Lessons learned in bullet points},with each point separated by a hyphen `-`"  # noqa
            )
            self.summary_dic["opinion"] = self.llm(prompt)
            self.summary_counter += 1

        @broadcast_service.on_listen("PaperSummaryTool.run.get_keywords")
        def get_keywords():
            prompt = (
                f"Please list no more than 7 keywords in the paper for the following abstract```{paper_summary}```"  # noqa
                "Your output format is:\nKeywords: keyword1, keyword2, keyword3"
            )
            self.summary_dic["keywords"] = self.llm(prompt)
            self.summary_counter += 1

        @broadcast_service.on_listen("PaperSummaryTool.run.get_advice")
        def get_advice():
            prompt = (
                f"Based on the abstract of the paper below, please provide 3-5 suggestions related to its topics or potential future research directions. Your output should be provided in bullet points  ```{paper_summary}```"  # noqa
                "Please use the following output format:\nRelated suggestions:\n{related suggestions in bullet points}, with each point separated by a hyphen `-`"  # noqa
            )
            self.summary_dic["advice"] = self.llm(prompt)
            self.summary_counter += 1

        self.summary_counter = 0
        # judge arxiv id or string type paper title
        if re.match(r"\d{4}\.\d{5}(v\d+)?", query):
            paper_info = self.arxiv_apiwrapper.query(
                id_list=[query], num_results=1, specified_fields=["title", "summary"]
            )
            if paper_info:
                paper_info = paper_info[0]
                paper_info["abstract"] = paper_info["summary"]
        else:
            try:
                paper_info = self.semantic_scholar_query_tool.run(
                    query,
                    return_type="original",
                    specified_fields=["title", "url", "abstract"],
                )
                if paper_info:
                    paper_info = paper_info[0]

            except NetWorkError:
                paper_info = self.arxiv_apiwrapper.query(
                    keyword=query, num_results=1, specified_fields=["title", "summary"]
                )
                if paper_info:
                    paper_info = paper_info[0]
                    paper_info["abstract"] = paper_info["summary"]

        if not paper_info:
            return (
                "semantic scholar query tool and arxiv query tool query result is None."
            )

        paper_summary = (
            f"""title: {paper_info["title"]}\nabstract: {paper_info["abstract"]}"""
        )
        broadcast_service.publish(
            "PaperSummaryTool.run.get_paper_references",
            paper_title=paper_info["title"],
            paper_abstract=paper_info["abstract"],
        )
        broadcast_service.publish("PaperSummaryTool.run.get_translate")
        time.sleep(0.01)
        broadcast_service.publish("PaperSummaryTool.run.get_opinion")
        time.sleep(0.01)
        broadcast_service.publish("PaperSummaryTool.run.get_advice")
        time.sleep(0.01)
        broadcast_service.publish("PaperSummaryTool.run.get_keywords")

        while self.summary_counter < 5:
            time.sleep(0.1)

        return (
            f"""{self.summary_dic["summary_zh"]}\n\n"""
            f"""{self.summary_dic["keywords"]}\n\n"""
            f"""{self.summary_dic["opinion"]}\n\n"""
            f"""{self.summary_dic["advice"]}\n\n"""
            f"""{self.summary_dic["references"]}"""
        )



import json
import warnings
from abc import ABC
from typing import Any, Dict, List, Optional

import requests

from promptulate.config import pne_config
from promptulate.error import LLMError
from promptulate.llms import BaseLLM
from promptulate.preset_roles.prompt import PRESET_SYSTEM_PROMPT_ERNIE
from promptulate.schema import AssistantMessage, LLMType, MessageSet, UserMessage
from promptulate.utils.logger import logger


class ErnieBot(BaseLLM, ABC):
    llm_type: LLMType = LLMType.ErnieBot
    """Used to MessageSet data convert"""
    model: str = "ernie-bot-4"
    """Model name to use."""
    temperature: float = 0.7
    """What sampling temperature to use."""
    top_p: float = 0.8
    """variety of text"""
    stream: bool = False
    """Whether to stream the results or not."""
    disable_search: bool = True
    """disable baidu search"""
    default_system_prompt: str = ""
    """ernie-bot system message"""
    enable_default_system_prompt: bool = True
    """enable use preset description"""
    penalty_score: float = 1.0
    system: str = ""
    api_param_keys: List[str] = [
        "temperature",
        "top_p",
        "stream",
        "stop",
        "system",
        "disable_search",
        "penalty_score",
    ]
    base_url: str = ""
    """set your base_url"""

    def __call__(
        self, instruction: str, stop: Optional[List[str]] = None, *args, **kwargs
    ) -> str:
        preset = (
            self.default_system_prompt
            if self.default_system_prompt != ""
            else PRESET_SYSTEM_PROMPT_ERNIE
        )
        if not self.enable_default_system_prompt:
            preset = ""
        self.system = preset
        message_set = MessageSet(
            messages=[
                UserMessage(content=instruction),
            ]
        )
        return self.predict(message_set, stop).content

    def _predict(
        self, prompts: MessageSet, stop: Optional[List[str]] = None, *args, **kwargs
    ) -> AssistantMessage:
        """
        Predicts the response using the ERNIE model.

        Args:
            prompts (MessageSet): The set of prompts to generate a response.
            stop (Optional[List[str]], optional): List of stop words to stop the
            generation. Defaults to None.

        Returns:
            AssistantMessage: The generated response.
        """
        warnings.warn(
            "QianFan class is online ,this module will deprecated", DeprecationWarning
        )
        headers = {"Content-Type": "application/json"}
        models = {
            "ernie-bot-turbo": pne_config.ernie_bot_turbo_url,
            "ernie-bot-4": pne_config.ernie_bot_4_url,
            "ernie-bot": pne_config.ernie_bot_url,
        }

        if self.base_url:
            url = self.base_url
        else:
            if models.__contains__(self.model):
                url = models[self.model]
            else:
                raise ValueError("pne not found this model")

        body: Dict[str, Any] = self._build_api_params_dict(prompts, stop)
        response = requests.post(
            url=url + "?access_token=" + pne_config.get_ernie_token(),
            headers=headers,
            json=body,
            proxies=pne_config.proxies,
        )
        if response.status_code == 200:
            # todo enable stream mode
            # for chunk in response.iter_content(chunk_size=None):
            #     logger.debug(chunk)
            ret_data = response.json()

            if ret_data.get("error_code", None):
                raise LLMError(ret_data)

            content: str = ret_data["result"]
            """ernie-bot official support stop,deprecated"""
            """if stop:
                    length: int = 1000000  # very large integer +inf
                    temp: str = ""
                    for s in stop:
                        temp_len = len(content.split(s)[0])
                        if temp_len < length:
                            temp = content.split(s)[0]
                            length = temp_len
                    content = temp"""
            logger.debug(f"[pne ernie answer] {content}")
            return AssistantMessage(content=content, additional_kwargs=ret_data)

    def _build_api_params_dict(
        self,
        prompts: MessageSet,
        stop: Optional[List[str]] = None,
    ) -> Dict[str, Any]:
        """Build api parameters to put it inside the body."""
        # print(prompts.type_)
        dic = {
            "messages": prompts.to_llm_prompt(self.llm_type),
        }

        if stop:
            dic.update({"stop": stop})

        for key in self.api_param_keys:
            if key in self.__dict__:
                dic.update({key: self.__dict__[key]})
        return dic



import re
from typing import Any, List

from typing_extensions import Literal


def _get_jinja2_variables(template: str) -> List[str]:
    try:
        from jinja2 import Environment, meta
    except ImportError:
        raise ImportError(
            "jinja2 not installed, which is needed to use the jinja2_formatter. "
            "Please install it with `pip install jinja2`."
        )
    env = Environment()
    ast = env.parse(template)
    variables = meta.find_undeclared_variables(ast)
    return list(variables)


def _jinja2_format(template: str, **kwargs) -> str:
    """Format a template using jinja2.

    *Security warning*: As of LangChain 0.0.329, this method uses Jinja2's
        SandboxedEnvironment by default. However, this sand-boxing should
        be treated as a best-effort approach rather than a guarantee of security.
        Do not accept jinja2 templates from untrusted sources as they may lead
        to arbitrary Python code execution.

        https://jinja.palletsprojects.com/en/3.1.x/sandbox/
    """
    try:
        from jinja2.sandbox import SandboxedEnvironment
    except ImportError:
        raise ImportError(
            "jinja2 not installed, which is needed to use the jinja2_formatter. "
            "Please install it with `pip install jinja2`."
            "Please be cautious when using jinja2 templates. "
            "Do not expand jinja2 templates using unverified or user-controlled "
            "inputs as that can result in arbitrary Python code execution."
        )

    # This uses a sandboxed environment to prevent arbitrary code execution.
    # Jinja2 uses an opt-out rather than opt-in approach for sand-boxing.
    # Please treat this sand-boxing as a best-effort approach rather than
    # a guarantee of security.
    # We recommend to never use jinja2 templates with untrusted inputs.
    # https://jinja.palletsprojects.com/en/3.1.x/sandbox/
    # approach not a guarantee of security.
    return SandboxedEnvironment().from_string(template).render(**kwargs)


class StringTemplate:
    """String Template for llm. It can generate a complex prompt."""

    def __init__(
        self, template: str, template_format: Literal["f-string", "jinja2"] = "f-string"
    ):
        self.template: str = template
        self.template_format: str = template_format
        self.variables: List[str] = []

        if template_format == "f-string":
            self.variables = re.findall(r"\{(\w+)\}", self.template)
        elif template_format == "jinja2":
            self.variables = _get_jinja2_variables(template)
        else:
            raise ValueError(
                f"template_format must be one of 'f-string' or 'jinja2'. Got: {template_format}"  # noqa: E501
            )

    def format(self, params: List[Any] = None, **kwargs) -> str:
        """Enter variables and return the formatted string."""
        if params:
            kwargs = {}
            for i, param in enumerate(params):
                kwargs.update({self.variables[i]: param})

        if self.template_format == "f-string":
            return self.template.format(**kwargs)
        elif self.template_format == "jinja2":
            return _jinja2_format(self.template, **kwargs)
